import pyrosetta
from functools import partial

def make_option_string(**options):
    """
    This just converts the key:value pairs to a command line string for the pyrosetta init.

    >> make_option_string(no_optH=False, ex1=None, remodel=dict(blueprint='mod.blue'))

    Bools are converted,
    None results in a value argument,
    Tuples are converted to xx:xx type arguments
    Dictionaries are converted to xx:xx type arguments (multiple, if multiple keys in the nested dictionary)

    Also... Full option list: https://www.rosettacommons.org/docs/latest/full-options-list
    """

    def format_inner(v):
        if v is None:
            return ''
        elif v in (False, True):
            return ' ' + str(v).lower()
        else:
            return ' ' + str(v)

    def format_outer(k):
        if isinstance(options[k], tuple):
            return '-' + ':'.join(k) + format_inner(options[k])
        elif isinstance(options[k], dict):
            return ' '.join([f'-{k}:{k2}' + format_inner(v2) for k2, v2 in options[k].items()])
        else:
            return '-' + k + format_inner(options[k])

    args = [format_outer(k) for k in options]
    return ' '.join(args)

# values many need changing.
common_options = dict(no_optH=False,
                      mute='All',
                      ex1=None, ex2=None,
                      ignore_unrecognized_res=True,
                      load_PDB_components=False,
                      ignore_waters=True)



import pyrosetta

Residue = pyrosetta.rosetta.core.conformation.Residue  # typehinting only


def is_xlinked(residue: Residue) -> bool:
    return get_xlink_idx(residue, raise_on_bond=False) != 0


def get_xlink_idx(residue: Residue, raise_on_bond: bool = True) -> int:
    cxi: int = sum((residue.has_lower_connect(), residue.has_upper_connect())) + 1
    if residue.n_current_residue_connections() >= cxi:
        return cxi
    elif raise_on_bond:
        raise ValueError('Not a crosslinked residue')
    else:
        return 0


def get_xlink_details(residue_i: int, pose: pyrosetta.Pose) -> dict:
    """
    cross-link (SSBOND, LINK etc.) => conn3
    By conn3 I mean a connection that is not
    lower (own N) or upper (own C).
    So if a residue has no connection on one or both of these,
    technically it will be conn2 or conn1, but that's being pedantic.

    NB. Does not check for incomplete connections (which are a product of fancy meddling)

    return
    :param residue_i:
    :param pose:
    :return: dict of keys other_idx, other_name3, other_is_protein own_atom_name other_atom_name
    """
    residue: Residue = pose.residue(residue_i)
    cxi: int = get_xlink_idx(residue)
    other_idx: int = residue.connected_residue_at_resconn(cxi)
    other: Residue = pose.residue(other_idx)
    atom_idx: int = residue.residue_connect_atom_index(cxi)
    # there **will** be a conn: (unless residue.has_incomplete_connection(3) == True)
    other_cxi: int = other.connections_to_residue(residue_i)[1]
    other_atom_idx: int = residue.residue_connect_atom_index(cxi)
    return dict(other_idx=other_idx,
                other_name3=other.name3(),
                other_is_protein=other.is_protein(),
                own_atom_name=residue.atom_name(atom_idx).strip(),
                other_atom_name=other.atom_name(other_atom_idx).strip(),
                )



__all__ = ['pose_from_file',
           'pose2pandas',
           'make_blank_pose',
           'add_bfactor_from_score',
           'get_last_res_in_chain',
           'clarify_selector',
           'count_ligands',
           'correct_numbering',
           'get_pdbstr',
           'pose_range']

from collections import Counter
from typing import (Optional, Tuple, Union, Iterable, Counter, List)

import numpy as np
import pandas as pd
import pyrosetta


def pose_from_file(pdb_filename: str,
                   params_filenames: Optional[Union[pyrosetta.rosetta.utility.vector1_string, List[str]]] = None) \
        -> pyrosetta.Pose:
    """
    Return a pose like pose_from_file but with params.

    :param pdb_filename:
    :param params_filenames:
    :return:
    """
    pose = make_blank_pose(params_filenames)
    pyrosetta.rosetta.core.import_pose.pose_from_file(pose, pdb_filename)
    return pose


vector1_string = pyrosetta.rosetta.utility.vector1_string


def make_blank_pose(params_filenames: Optional[Union[vector1_string, List[str]]] = None) \
        -> pyrosetta.Pose:
    """
    Returns an empty pose, but with params from files.

    :param params_filenames:
    :return:
    """
    pose = pyrosetta.Pose()
    if params_filenames and isinstance(params_filenames, vector1_string):
        pyrosetta.generate_nonstandard_residue_set(pose, params_filenames)
    if params_filenames and isinstance(params_filenames, list):
        params_filenames2 = vector1_string()
        params_filenames2.extend([f for f in params_filenames if f])
        pyrosetta.generate_nonstandard_residue_set(pose, params_filenames2)
    elif params_filenames:
        raise TypeError(f'Unexpected type {type(params_filenames)}')
    else:
        pass
    return pose


def pose2pandas(pose: pyrosetta.Pose, scorefxn: pyrosetta.ScoreFunction) -> pd.DataFrame:
    """
    Return a pandas dataframe from the scores of the pose

    :param pose:
    :return:
    """
    pose.energies().clear_energies()
    scorefxn.weights() # neccessary?
    emopts = pyrosetta.rosetta.core.scoring.methods.EnergyMethodOptions(scorefxn.energy_method_options())
    emopts.hbond_options().decompose_bb_hb_into_pair_energies(True)
    scorefxn.set_energy_method_options(emopts)
    scorefxn(pose)
    scores = pd.DataFrame(pose.energies().residue_total_energies_array())
    pi = pose.pdb_info()
    scores['residue'] = scores.index.to_series() \
        .apply(lambda r: pose.residue(r + 1) \
               .name1() + pi.pose2pdb(r + 1)
               )
    return scores


def add_bfactor_from_score(pose: pyrosetta.Pose):
    """
    Adds the bfactors from total_score.
    Snippet for testing in Jupyter

    >>> import nglview as nv
    >>> view = nv.show_rosetta(pose)
    >>> # view = nv.show_file('test.cif')
    >>> view.clear_representations()
    >>> view.add_tube(radiusType="bfactor", color="bfactor", radiusScale=0.10, colorScale="RdYlBu")
    >>> view

    ``replace_res_remap_bfactors`` may have been a cleaner strategy. This was quicker to write.

    If this fails, it may be because the pose was not scored first.
    """
    if pose.pdb_info().obsolete():
        raise ValueError('Pose pdb_info is flagged as obsolete (change `pose.pdb_info().obsolete(False)`)')
    # scores
    energies = pose.energies()

    def get_res_score(res):
        total_score = pyrosetta.rosetta.core.scoring.ScoreType.total_score
        # if pose.residue(res).is_polymer()
        try:
            return energies.residue_total_energies(res)[total_score]
        except:
            return float('nan')

    # the array goes from zero (nan) to n_residues
    total_scores = np.array([float('nan')] + [get_res_score(res) for res in range(1, pose.total_residue() + 1)])
    mask = np.isnan(total_scores)
    total_scores -= np.nanmin(total_scores)
    total_scores *= 100 / np.nanmax(total_scores)
    total_scores = np.nan_to_num(total_scores, nan=100)
    total_scores[mask] = 0.
    # add to pose
    pdb_info = pose.pdb_info()
    for res in range(pose.total_residue()):
        for i in range(pose.residue(res + 1).natoms()):
            pdb_info.bfactor(res + 1, i + 1, total_scores[res + 1])


def get_last_res_in_chain(pose, chain='A') -> int:
    """
    Returns last residue index in a chain. THere is probably a mover that does this.

    :param pose:
    :param chain: letter or number
    :return:
    """
    cv = pyrosetta.rosetta.core.select.residue_selector.ChainSelector(chain).apply(pose)
    rv = pyrosetta.rosetta.core.select.residue_selector.ResidueVector(cv)
    return max(rv)


def clarify_selector(selector: pyrosetta.rosetta.core.select.residue_selector.ResidueSelector,
                     pose: pyrosetta.Pose) -> List['str']:
    """
    Given a selector and pose return a list of residues in NGL selection format
    Example, [CMP]787:H

    :param selector:
    :param pose:
    :return: list of residues in NGL selection format
    """
    pose2pdb = pose.pdb_info().pose2pdb
    vector = selector.apply(pose)
    rv = pyrosetta.rosetta.core.select.residue_selector.ResidueVector(vector)
    return [f'[{pose.residue(r).name3()}]{pose2pdb(r).strip().replace(" ", ":")}' for r in rv]

def count_ligands(pose: pyrosetta.Pose) -> List[Tuple[str, int]]:
    LIGAND = pyrosetta.rosetta.core.chemical.ResidueProperty.LIGAND
    pr_rs = pyrosetta.rosetta.core.select.residue_selector
    lig_sele = pr_rs.ResiduePropertySelector(LIGAND)
    count = Counter([pose.residue(i).name3() for i in pr_rs.ResidueVector(lig_sele.apply(pose))])
    return count.most_common()

def correct_numbering(pose):
    """
    A fresh PDBInfo has the PDB residue number the same as the pose one
    as opposed to restarting per chain.

    :param pose:
    :return:
    """
    pdb_info = pose.pdb_info()
    if pdb_info is None:
        pdb_mover = pyrosetta.rosetta.protocols.simple_moves.AddPDBInfoMover()
        pdb_mover.apply(pose)
        pdb_info = pose.pdb_info()
    current_chain = 'A'
    r = 1
    for i in range(1, pose.total_residue() +1):
        if pdb_info.chain(i) != current_chain:
            current_chain = pdb_info.chain(i)
            r = 1
            pdb_info.number(i, 1)
        else:
            pdb_info.number(i, r)
            r += 1

def get_pdbstr(pose):
    buffer = pyrosetta.rosetta.std.stringbuf()
    pose.dump_pdb(pyrosetta.rosetta.std.ostream(buffer))
    return buffer.str()


def pose_range(pose: pyrosetta.Pose, protein_only=True) -> Iterable:
    """
    range but for the pose...
    For now naive of chains, non-amino acid residues etc.
    """
    all_iter = range(1, pose.total_residue()+1)
    if protein_only:
        return filter(lambda i: pose.residue(i).is_protein(), all_iter)
    return all_iter




import pyrosetta
import requests
import warnings

from ..common_ops.utils import make_blank_pose

try:
    from rdkit_to_params import Params, neutralise
except ImportError:
    from warnings import warn
    warn("rdkit_to_params couldn't be imported."
         " LiganNicker and parameterised_pose_from_file won't work.")
    Params = None
    pass
from typing import (Union, Dict, List)

__all__ = ['parameterised_pose_from_file', 'parameterised_pose_from_pdbblock', 'get_smiles']


def parameterised_pose_from_file(pdb_filename,
                                 wanted_ligands: Union[List[str], Dict[str, Union[str, None]]] = (),
                                 force_parameterisation: bool = False,
                                 neutralise_params: bool = True,
                                 save_params: bool = True,
                                 overriding_params=()) -> pyrosetta.Pose:
    """
    pose loading, the circutous way to not loose ligand or use PDB_component.
    Assumes all mystery components are PDB residues.
    Works best with ignore_unrecognized_res False

    :param pdb_filename:
    :param wanted_ligands: a list of three letter codes or a dictionary of three letter codes to None or smiles
    :param force_parameterisation:
    :param neutralise_params: protonated for pH 7
    :param save_params:
    :param overriding_params: list of params filenames
    :return:
    """
    with open(pdb_filename, 'r') as fh:
        pdbblock = fh.read()
    return parameterised_pose_from_pdbblock(pdbblock,
                                            wanted_ligands=wanted_ligands,
                                            force_parameterisation=force_parameterisation,
                                            neutralise_params=neutralise_params,
                                            save_params=save_params,
                                            overriding_params=overriding_params)


def parameterised_pose_from_pdbblock(pdbblock: str,
                                     wanted_ligands: Union[List[str], Dict[str, Union[str, None]]] = (),
                                     force_parameterisation: bool = False,
                                     neutralise_params: bool = True,
                                     save_params: bool = True,
                                     overriding_params=()) -> pyrosetta.Pose:
    """
    pose loading, the circutous way to not loose ligand or use PDB_component.
    Assumes all mystery components are PDB residues.
    Works best with ignore_unrecognized_res False

    :param pdb_filename:
    :param wanted_ligands: a list of three letter codes or a dictionary of three letter codes to None or smiles
    :param force_parameterisation:
    :param neutralise_params: protonated for pH 7
    :param save_params:
    :param overriding_params: list of params filenames
    :return:
    """

    if pyrosetta.rosetta.basic.options.get_boolean_option('in:file:load_PDB_components'):
        raise ValueError('load_PDB_components is True. Run ``pyrosetta.pose_from_filename`` then.')
    # check if ignore_unrecognized_res ?
    if isinstance(wanted_ligands, dict):
        needed_ligands = dict(wanted_ligands)
    else:
        needed_ligands = {lig: None for lig in wanted_ligands}
    pose = _prep_pose(pdbblock=pdbblock,
                      ligand_dex=needed_ligands,
                      force_parameterisation=force_parameterisation,
                      neutralise_params=neutralise_params,
                      save_params=save_params,
                      overriding_params=overriding_params)
    assert pose.sequence(), 'The structure failed to load'
    return pose


def _prep_pose(pdbblock: str,
               ligand_dex: Dict,
               force_parameterisation: bool = False,
               neutralise_params: bool = True,
               save_params: bool = True,
               overriding_params: List = ()):
    """
    Parameterise if needed the needed_ligands and add to the list if there's an additional missing PDB residue.
    """
    pose = make_blank_pose(overriding_params)
    rts = pose.residue_type_set_for_pose()
    try:
        for target_ligand in ligand_dex:
            if not rts.has_name3(target_ligand) or force_parameterisation:
                smiles = ligand_dex[target_ligand]
                if not smiles:
                    smiles = get_smiles(target_ligand)
                params = parameterise(pdb_block=pdbblock,
                                      target_ligand=target_ligand,
                                      smiles=smiles,
                                      neutral=neutralise_params,
                                      save=save_params)
                rts = params.add_residuetype(pose)
                pose.conformation().reset_residue_type_set_for_conf(rts)
                rts = pose.residue_type_set_for_pose()  # playing it ubersafe
                assert rts.has_name3(target_ligand)
        pyrosetta.rosetta.core.import_pose.pose_from_pdbstring(pose, pdbblock)
        # pyrosetta.pose_from_file(pose, pdb_filename)
        return pose
    except RuntimeError as err:
        missing = err.args[0].strip()[-3:]
        warnings.warn(f'adding {missing}')
        ligand_dex[missing] = None
        return _prep_pose(pdbblock, ligand_dex, force_parameterisation, neutralise_params, save_params,
                          overriding_params)


def parameterise(pdb_block: str, target_ligand: str, smiles: str, neutral: bool = True, save: bool = True) -> Params:
    params = Params.from_smiles_w_pdbblock(pdb_block=pdb_block,
                                           smiles=smiles,
                                           proximityBonding=False,
                                           name=target_ligand)
    if neutral:
        mol = neutralise(params.mol)
        params = Params.from_mol(mol, name=target_ligand)
    if save:
        params.dump(f'{target_ligand}.params')  # safekeeping. Not used.
    return params


def get_smiles(ligand_code: str) -> str:
    """
    Get the smiles of a ligand.
    Remember that PDBe smiles need to charged to pH 7.
    """
    ligand_code = ligand_code.upper()
    ligand_data = requests.get(f'https://www.ebi.ac.uk/pdbe/api/pdb/compound/summary/{ligand_code}').json()
    return ligand_data[ligand_code][0]['smiles'][0]['name']



import pyrosetta
import re


class Mutation:
    """
    A mutation is an object that has all the details of the mutation.
    A variant, as interpreted in ``Model.score_mutations`` is a pose with a mutation.
    """
    _name3 = {'A': 'ALA',
              'C': 'CYS',
              'D': 'ASP',
              'E': 'GLU',
              'F': 'PHE',
              'G': 'GLY',
              'H': 'HIS',
              'I': 'ILE',
              'L': 'LEU',
              'K': 'LYS',
              'M': 'MET',
              'N': 'ASN',
              'P': 'PRO',
              'Q': 'GLN',
              'R': 'ARG',
              'S': 'SER',
              'T': 'THR',
              'V': 'VAL',
              'W': 'TRP',
              'Y': 'TYR'}

    def __init__(self, mutation_name: str, chain: str, pose: pyrosetta.Pose):
        self.mutation = self.parse_mutation(mutation_name)
        rex = re.match('(\w)(\d+)(\w)', self.mutation)
        self.pdb_resi = int(rex.group(2))
        self.chain = chain
        self.from_resn1 = rex.group(1)
        self.from_resn3 = self._name3[rex.group(1)]
        self.to_resn1 = rex.group(3)
        self.to_resn3 = self._name3[rex.group(3)]
        pose2pdb = pose.pdb_info().pdb2pose
        self.pose_resi = pose2pdb(res=self.pdb_resi, chain=self.chain)
        if self.pose_resi != 0:
            self.pose_residue = pose.residue(self.pose_resi)
            self.pose_resn1 = self.pose_residue.name1()
            self.pose_resn3 = self.pose_residue.name3()
        else:
            self.pose_residue = None
            self.pose_resn1 = None
            self.pose_resn3 = None

    def parse_mutation(self, mutation: str):
        if mutation[:2] == 'p.':
            mutation = mutation.replace('p.', '')
        if mutation[1].isdigit():
            return mutation
        else:
            value2key = lambda value: list(self._name3.keys())[list(self._name3.values()).index(value.upper())]
            return value2key(mutation[:3]) + mutation[3:-3] + value2key(mutation[-3:])

    def is_valid(self):
        return self.pose_resn1 == self.from_resn1

    def assert_valid(self):
        assert self.is_valid(), f'residue {self.pose_resi}(pose)/{self.pdb_resi}(pdb) ' + \
                                f'is a {self.pose_resn3}, not a {self.from_resn3}'

    def __str__(self):
        return self.mutation


from typing import List
import pyrosetta

class BlueprinterExpected:
    def expected_seq(self):
        """

        :return:
        """
        return ''.join([self.get_expected_aa_from_row(row) for row in self])

    def get_expected_aa_from_row(self, row: List[str]) -> str:
        if len(row) == 0: # impossible
            return ''
        elif len(row) == 3 and row[2] == '.': # "30 K ."
            return row[1]
        elif len(row) == 3:  # "30 K H" ???
            return '*'  # H/L/E/D in lowercase would be ambiguous
        elif len(row) == 4 and row[3] in ('NATAA', 'NATRO'):  # "30 K H NATAA"
            return row[1]
        elif len(row) == 4 and 'PIKAA' in row[3] and len(row[3]) == len('PIKAA ')+1:  # "30 K H PIKAA R"
            return row[3][-1]
        else: # ALLAA etc..
            return '*'

    def show_aligned(self, pose):
        """
        notebook output alignment of pose seq and blueprint expectation
        :param pose:
        :return:
        """
        self.show_seq_aligned(pose.sequence(), self.expected_seq())

    def _show_alignment(self, alignment: list):
        from Bio import pairwise2
        from IPython.display import display, HTML
        formatted = pairwise2.format_alignment(*alignment)
        a, gap, b, score = formatted.strip().split('\n')
        gap = ''.join(['.' if c == '|' else '*' for c in gap])
        display(HTML(f'<div style="font-family:monospace; display: inline-block; white-space: nowrap;">'+
                     f'{a}<br/>{gap.replace(" ", "*")}<br/>{b}<br/>{score}</div>'))
        #ignore pycharm. typehint for display is wrong.

    def show_poses_aligned(self, pose_A, pose_B):
        self.show_seq_aligned(pose_A.sequence(), pose_B.sequence())

    def show_seq_aligned(self, seq_A: str, seq_B:str):
        from Bio import pairwise2
        alignment = pairwise2.align.globalxx(seq_A, seq_B)[0]
        self._show_alignment(alignment)
        return alignment

    def correct(self, pose: pyrosetta.Pose) -> pyrosetta.rosetta.core.select.residue_selector.ResidueSelector:
        """
        This is not a great thing to do. So it is best to relax the neighbours of the vector afterwards.
        If there are valines instead of the intended sequence that is bad.
        Altering the blueprint is required.

        :param pose:
        :return:
        """
        MutateResidue = pyrosetta.rosetta.protocols.simple_moves.MutateResidue
        one2three = pyrosetta.rosetta.protocols.motifs.name3_from_oneletter
        ex_seq = self.expected_seq()
        altered = pyrosetta.rosetta.core.select.residue_selector.ResidueIndexSelector()
        for i, (expected, current) in enumerate(zip(ex_seq, pose.chain_sequence(1))):
            if expected != current:
                altered.append_index(i + 1)
                MutateResidue(target=i + 1, new_res=one2three(expected)).apply(pose)
        return altered

    def correct_and_relax(self, pose: pyrosetta.Pose):
        altered = self.correct(pose)
        movemap = pyrosetta.MoveMap()
        NeighborhoodResidueSelector = pyrosetta.rosetta.core.select.residue_selector.NeighborhoodResidueSelector
        neigh_sele = NeighborhoodResidueSelector(altered, distance=15, include_focus_in_subset=True)
        n = neigh_sele.apply(pose)
        movemap.set_bb(allow_bb=n)
        movemap.set_chi(allow_chi=n)
        scorefxn = pyrosetta.get_fa_scorefxn()
        # cartesian
        scorefxn_cart = pyrosetta.create_score_function('ref2015_cart')
        relax = pyrosetta.rosetta.protocols.relax.FastRelax(scorefxn_cart, 5)
        relax.cartesian(True)
        relax.minimize_bond_angles(True)
        relax.minimize_bond_lengths(True)
        relax.set_movemap_disables_packing_of_fixed_chi_positions(True)
        relax.set_movemap(movemap)
        relax.apply(pose)
        # dual
        relax = pyrosetta.rosetta.protocols.relax.FastRelax(scorefxn, 5)
        relax.set_movemap_disables_packing_of_fixed_chi_positions(True)
        relax.set_movemap(movemap)
        relax.apply(pose)








__all__ = ['AF2NotebookAnalyser']

import json
import os
import re
from typing import (Optional, Dict)

import numpy as np
import pandas as pd
import pyrosetta

pr_rs = pyrosetta.rosetta.core.select.residue_selector
from .retrieval import reshape_errors
from .constraints import add_pae_constraints, add_interchain_pae_constraints
from ..common_ops import pose_from_file
import pickle
from collections import defaultdict


class AF2NotebookAnalyser:

    def __init__(self, folder: str, load_poses: bool = True):
        self.folder = folder
        self.scores = self.make_AF2_dataframe()
        self._add_settings()
        self.relaxed_poses = dict()
        self.original_poses = dict()
        self.phospho_poses = dict()
        self.pose_groupnames = ['relaxed', 'original', 'phospho']
        if load_poses:
            self.original_poses = self.get_poses()
        self.errors = self.get_errors()

    @property
    def poses(self):
        # to alter change pose_groupnames and add the required *_poses attributes, I guess.
        return {groupname: getattr(self, groupname + '_poses') for groupname in self.pose_groupnames}

    def make_AF2_dataframe(self) -> pd.DataFrame:
        """
        Given a folder form ColabsFold return a dictionary
        with key rank index
        and value a dictionary of details

        This is convoluted, but it may have been altered by a human.
        """
        filenames = os.listdir(self.folder)
        # group files
        ranked_filenames = {}
        for filename in filenames:
            if not re.match('rank_\d+.*\.pdb', filename):
                continue
            rex = re.match('rank_(\d+)_model_(\d+).*seed_(\d+)(.*)\.pdb', filename)
            rank = int(rex.group(1))
            model = int(rex.group(2))
            seed = int(rex.group(3))
            other = rex.group(4)
            if rank in ranked_filenames and ranked_filenames[rank]['relaxed'] == True:
                continue
            data = dict(name=filename,
                        path=os.path.join(self.folder, filename),
                        rank=rank,
                        model=model,
                        seed=seed,
                        relaxed='_relaxed' in other
                        )
            ranked_filenames[rank] = data
        # make dataframe
        return pd.DataFrame(list(ranked_filenames.values()))

    def _add_settings(self):
        # add data from settings.
        pLDDTs = {}
        pTMscores = {}
        with open(os.path.join(self.folder, 'settings.txt'), 'r') as fh:
            for match in re.findall('rank_(\d+).*pLDDT\:(\d+\.\d+)\ pTMscore:(\d+\.\d+)', fh.read()):
                pLDDTs[int(match[0])] = float(match[1])
                pTMscores[int(match[0])] = float(match[2])
        self.scores['pLDDT'] = self.scores['rank'].map(pLDDTs)
        self.scores['pTMscore'] = self.scores['rank'].map(pTMscores)

    def get_poses(self) -> Dict[int, pyrosetta.Pose]:
        """
        This used to be a ``pyrosetta.rosetta.utility.vector1_core_pose_Pose``
        but it turns out that the getitem of ``vector1_core_pose_Pose``
        returns a clone of the pose, not the pose itself.

        :return:
        """
        # poses = pyrosetta.rosetta.utility.vector1_core_pose_Pose(len(self.scores))
        poses = dict()
        for i, row in self.scores.iterrows():
            poses[row['rank']] = pyrosetta.pose_from_file(row['path'])
        return poses

    def get_errors(self) -> dict:
        errors = dict()
        for i, row in self.scores.iterrows():
            filename = row['path'].replace('.pdb', '.json') \
                .replace('_unrelaxed', '_pae') \
                .replace('_relaxed', '_pae')
            with open(filename, 'r') as fh:
                errors[row['rank']] = reshape_errors(json.load(fh))
        return errors

    def _generator_poses(self, groupname: str = 'relaxed'):
        """
        :return: rank integer, pose, error dictionary
        """
        assert len(self.poses[groupname]), f'The group {groupname} is not loaded'
        return ((index, self.poses[groupname][index], self.errors[index]) for index in self.errors)

    def constrain(self, groupname: str = 'relaxed', **add_pae_constraints_arguments):
        if len(self.original_poses) == 0:
            raise ValueError('Load poses first.')
        for index, pose, error in self._generator_poses(groupname):
            add_pae_constraints(pose, error, **add_pae_constraints_arguments)
            add_interchain_pae_constraints(pose, error, cutoff=15)

    def sidechain_relax(self, cycles: int = 5):
        vanilla_scorefxn = pyrosetta.get_fa_scorefxn()
        ap_st = pyrosetta.rosetta.core.scoring.ScoreType.atom_pair_constraint
        vanilla_scorefxn.set_weight(ap_st, 0)
        movemap = pyrosetta.MoveMap()
        movemap.set_bb(False)
        movemap.set_chi(True)
        chirelax = pyrosetta.rosetta.protocols.relax.FastRelax(vanilla_scorefxn, cycles)
        chirelax.set_movemap(movemap)
        for index, pose, error in self._generator_poses('original'):
            if index not in self.relaxed_poses:
                self.relaxed_poses[index] = pose.clone()
            chirelax.apply(self.relaxed_poses[index])

    def relax(self, cycles: int = 3):
        if self.original_poses is None:
            raise ValueError('Load poses first.')
        scorefxn = pyrosetta.get_fa_scorefxn()
        ap_st = pyrosetta.rosetta.core.scoring.ScoreType.atom_pair_constraint
        scorefxn.set_weight(ap_st, 1)
        for index, pose, error in self._generator_poses('original'):
            if index not in self.relaxed_poses:
                self.relaxed_poses[index] = pose.clone()
            relax = pyrosetta.rosetta.protocols.relax.FastRelax(scorefxn, cycles)
            movemap = pyrosetta.MoveMap()
            movemap.set_bb(True)
            movemap.set_chi(True)
            movemap.set_jump(True)
            relax.set_movemap(movemap)
            relax.apply(self.relaxed_poses[index])
        # Add dG
        scorefxn.set_weight(ap_st, 0)
        self.scores['dG'] = self.scores['rank'].apply(lambda rank: scorefxn(self.relaxed_poses[rank]))

    def constrain_and_relax(self, cycles: int = 3):
        self.sidechain_relax()
        self.constrain()
        self.relax(cycles)

    def calculate_interface(self, interface='A_B'):
        self.get_median_interface_bfactors()
        # interface score.
        if 'dG' not in self.scores:
            return
        newdata = []
        for rank in self.scores['rank']:
            ia = pyrosetta.rosetta.protocols.analysis.InterfaceAnalyzerMover(interface)
            ia.apply(self.relaxed_poses[rank])
            newdata.append({'complex_energy':             ia.get_complex_energy(),
                            'separated_interface_energy': ia.get_separated_interface_energy(),
                            'complexed_sasa':             ia.get_complexed_sasa(),
                            'crossterm_interface_energy': ia.get_crossterm_interface_energy(),
                            'interface_dG':               ia.get_interface_dG(),
                            'interface_delta_sasa':       ia.get_interface_delta_sasa()})
        # adding multiple columns, hence why not apply route.
        # the order is not chanced, so all good
        newdata = pd.DataFrame(newdata)
        for column in newdata.columns:
            self.scores[column] = newdata[column]

    # ------------------------------------------------------------------------

    def _parse_folder_argument(self, folder):
        if folder and not os.path.exists(folder):
            # user specified non existant folder.
            os.mkdir(folder)
            return folder
        elif folder:
            return folder
        else:
            # use original folder.
            return self.folder

    def dump_pdbs(self,
                  groupname: str = 'relaxed',
                  folder: Optional[str] = None,
                  prefix: Optional[str] = ''):
        for index, pose, error in self._generator_poses(groupname):
            folder = self._parse_folder_argument(folder)
            path = os.path.join(folder, f'{prefix}rank_{index}_pyrosetta_{groupname}.pdb')
            pose.dump_pdb(path)

    def dump(self, folder: Optional[str] = None):
        folder = self._parse_folder_argument(folder)
        self.scores.to_csv(os.path.join(folder, 'scores.csv'))
        self.dump_pdbs(folder=folder)
        with open(os.path.join(folder, 'errors.p', 'w')) as fh:
            pickle.dump(self.errors, fh)
        for groupname in self.poses:
            if len(self.poses[groupname]) != 0:
                self.dump_pdbs(groupname=groupname, folder=folder)

    @classmethod
    def load(cls, folder: str, load_poses=False, params=()):
        self = cls(folder=folder, load_poses=load_poses)
        self.scores = pd.read_csv(os.path.join(folder, 'scores.csv'), index_col=0)
        valids = [filename for filename in os.listdir(folder) if '.pdb' in filename]
        ranker = lambda filename: int(re.search(r'rank(\d+)\.pdb', filename).group(1))
        self.relaxed_poses = {ranker(filename): pose_from_file(filename, params) for filename in valids}
        with open(os.path.join(folder, 'errors.p', 'r')) as fh:
            self.errors = pickle.load(fh)
        return self

    # ----------------------------------------------------

    def get_interactions(self,
                         pose: pyrosetta.Pose,
                         chain_id: int,
                         threshold: float = 3.) -> pr_rs.ResidueVector:
        chain_sele = pr_rs.ChainSelector(chain_id)
        other_chains_sele = pr_rs.NotResidueSelector(chain_sele)
        cc_sele = pr_rs.CloseContactResidueSelector()
        cc_sele.central_residue_group_selector(other_chains_sele)
        cc_sele.threshold(float(threshold))
        other_cc_sele = pr_rs.AndResidueSelector(chain_sele, cc_sele)
        return pr_rs.ResidueVector(other_cc_sele.apply(pose))

    def find_interface_residues(self):
        if 'N_interchain_residues_1' in self.scores:
            return  # already run
        assert self.relaxed_poses, 'No poses loaded yet!'
        assert self.relaxed_poses[1].num_chains() > 1, 'Single chain!'
        self.scores['interchain_residues_1'] = self.scores['rank'].apply(
            lambda rank: self.get_interactions(self.relaxed_poses[rank], 1))
        self.scores['interchain_residues_2'] = self.scores['rank'].apply(
            lambda rank: self.get_interactions(self.relaxed_poses[rank], 2))
        self.scores['N_interchain_residues_1'] = self.scores['interchain_residues_1'].apply(len)
        self.scores['N_interchain_residues_2'] = self.scores['interchain_residues_2'].apply(len)

    def _get_median_interface_bfactor(self, pose, residues):
        pbd_info = pose.pdb_info()
        bfactors = [pbd_info.bfactor(r, pose.residue(r).atom_index('CA')) for r in residues]
        return np.median(bfactors)

    def get_median_interface_bfactors(self):
        self.find_interface_residues()
        for i, row in self.scores.iterrows():
            residues = list(row.interchain_residues_1) + list(row.interchain_residues_2)
            pose = self.relaxed_poses[row['rank']]  # ``row.rank`` is a function, just like ``row.name``.
            return self._get_median_interface_bfactor(pose, residues)

    # -------------------------------------------------

    @classmethod
    def parse_phosphosite(cls, raw: str, minimum: int = 1, maximum: int = -1):
        """
        A rubbish method to convert copy-pasted Phosphosite web table.

        :param raw:
        :param minimum:
        :param maximum:
        :return:
        """
        if maximum < 1:
            maximum = float('nan')
        assert '-p' in raw, f'Is this what you wanted in your clipboard: {raw}'
        ptms = defaultdict(list)
        for resn, resi, mod in re.findall('(\w)(\d+)\-(\w+)', raw):
            if int(resi) < minimum or int(resi) > maximum:
                continue
            ptms[mod].append(int(resi))
        return ptms

    def _apply_patch(self, pose, patch, residues, restriction=None):
        MutateResidue = pyrosetta.rosetta.protocols.simple_moves.MutateResidue
        valids = []
        for r in residues:
            if r == 0:
                continue  # missing
            present_res = pose.residue(r).name3()
            new_res = f"{present_res.upper()}:{patch}"
            if restriction is not None and present_res.upper() != restriction.upper():
                continue
            MutateResidue(target=r, new_res=new_res).apply(pose)
            valids.append(r)
        return valids

    def _phospho_pose(self, pose, scorefxn, cycles=3, **modifications):
        valids = []
        for ptm, residues in modifications.items():
            if ptm == 'ub':
                continue  # What is a proxy for ubiquitination??
            elif ptm == 'p':
                v = self._apply_patch(pose,
                                      patch='phosphorylated',
                                      residues=residues)
                valids.extend(v)
            elif ptm == 'ac':
                v = self._apply_patch(pose,
                                      patch='acetylated',
                                      residues=residues)
                valids.extend(v)
            elif ptm == 'm1':
                v = self._apply_patch(pose,
                                      patch='monomethylated',
                                      residues=residues,
                                      restriction='LYS')  # monomethylarginine (NMM) will segfault
                valids.extend(v)
            elif ptm == 'm2':
                v = self._apply_patch(pose,
                                      patch='dimethylated',
                                      residues=residues,
                                      restriction='LYS')  # dimethylarginine (DA2) will segfault
                valids.extend(v)
            elif ptm == 'm3':
                v = self._apply_patch(pose,
                                      patch='trimethylated',
                                      residues=residues,
                                      restriction='LYS')  # is trimethylarginine a thing?
                valids.extend(v)
            else:
                continue  # no Gal
                # raise ValueError(f'What is {ptm}?')
        # minimise
        resi_sele = pyrosetta.rosetta.core.select.residue_selector.ResidueIndexSelector()
        for r in valids:
            resi_sele.append_index(r)
        if len(valids) == 0:
            return valids
        neigh_sele = pyrosetta.rosetta.core.select.residue_selector.NeighborhoodResidueSelector(resi_sele, 7, True)
        neighbour_vector = neigh_sele.apply(pose)
        relax = pyrosetta.rosetta.protocols.relax.FastRelax(scorefxn, cycles)
        movemap = pyrosetta.MoveMap()
        movemap.set_bb(neighbour_vector)
        movemap.set_chi(neighbour_vector)
        relax.set_movemap(movemap)
        relax.apply(pose)
        return valids

    def make_phosphorylated(self,
                            pdb_ptms: dict,
                            chain: str = 'A',
                            cycles: int = 3):
        # convert PTM to pose
        pdb_info = self.original_poses[1].pdb_info()
        convert = lambda pdb_residues: [pdb_info.pdb2pose(res=r, chain=chain) for r in pdb_residues]
        ptms = {k: convert(pdb_residues) for k, pdb_residues in pdb_ptms.items()}
        # scorefunction
        scorefxn = pyrosetta.get_fa_scorefxn()
        ap_st = pyrosetta.rosetta.core.scoring.ScoreType.atom_pair_constraint
        scorefxn.set_weight(ap_st, 1)
        for index, pose, error in self._generator_poses('relaxed'):
            self.phospho_poses[index] = pose.clone()
            self._phospho_pose(self.phospho_poses[index], scorefxn, cycles=cycles, **ptms)



__all__ = ['make_pae_plot']

import numpy as np

def make_pae_plot(errors: np.ndarray) -> 'plotly.import graph_objs.Figure':
    """
    Make AlphaFold2-EBI like PAE plot
    """
    import plotly.express as px
    fig = px.imshow(errors, color_continuous_scale=[(0, 'green'), (1, 'white')])
    return fig



from typing import (List)

import pyrosetta

from ..common_ops import pose_range


def get_cis_residues(pose: pyrosetta.Pose) -> List[int]:
    """
    Returns the pose indices of residues in cis (omega of zero)
    """
    # the cis residue is actually the one after (also a filter is not a generator)
    cis_gen = filter(lambda ri: abs(pose.omega(ri)) < 90., pose_range(pose))
    # citri is not really the plural of cis as cis is a preposition, citer is its adjective.
    # but best not discuss Latin grammar _ulterior_-ly.
    citri: List[int] = []
    for i in cis_gen:  #: int
        if i + 1 > pose.total_residue():
            # terminal residue
            continue
        if not pose.residue(i).connected_residue_at_upper():
            # gap
            continue
        citri.append(i + 1)
    return citri



from typing import Tuple, List, Union
import re
from .chain_ops import ChainOps


class Transmogrifier(ChainOps):
    """
    This is to convert a mutation from one species (``owned_label``) to another (``wanted_label``)
    based on provided sequences in the chains list of dict

    """

    def __init__(self, chains: List[dict], wanted_label: str, owned_label: str):
        super().__init__(chains)  # self.chains
        self.wanted_label = wanted_label
        self.owned_label = owned_label

    @classmethod
    def from_chain_ops(cls, chain_ops: ChainOps, wanted_label: str, owned_label: str):
        return cls(chain_ops.chains, wanted_label, owned_label)

    def align_seqs(self, chain_selection) -> Tuple[str, str]:
        """
        Align the human seq to the mouse and store it the chain dict
        """
        from Bio import pairwise2
        chain = self.chains[chain_selection]
        alignments = pairwise2.align.globalxs(chain[f'{self.wanted_label}_sequence'],
                                              chain[f'{self.owned_label}_sequence'],
                                              -1,  # open
                                              -0.1  # extend
                                              )
        al = alignments[0]
        chain[f'{self.wanted_label}_aln_sequence'] = al[0]
        chain[f'{self.owned_label}_aln_sequence'] = al[1]
        return al[0], al[1]

    def covert_A2B(self, seqA: str, seqB: str, resiA: int) -> int:
        """
        Given seqA and seqB as two gap aligned sequences,
        and an off-by-one residue number of seqA without counting gaps,
        return an off-by-one residue number of seqB without counting gaps.
        """
        assert resiA > 0, 'Negative number is no.'
        assert isinstance(resiA, int), 'Float is no'
        # first step
        get_resi2aln_map = lambda seq: [i for i, r in enumerate(seq) if r != '-']
        aln_pos = get_resi2aln_map(seqA)[resiA - 1]
        # second
        return get_resi2aln_map(seqB).index(aln_pos) + 1

    def transmogrify(self, mutation: str, chain_selection: Union[str, int, dict]) -> str:
        chain = self[chain_selection]
        human_resi = int(re.search(r'(\d+)', mutation).group(1))
        mouse_resi = self.covert_A2B(chain[f'{self.owned_label}_aln_sequence'],
                                     chain[f'{self.wanted_label}_aln_sequence'],
                                     human_resi)
        return re.sub(str(human_resi), str(mouse_resi), mutation)


class Murinizer(Transmogrifier):
    """
    Human --> Mouse
    """

    def __init__(self, chains: List[dict]):
        super().__init__(chains, 'mouse', 'human')

# from functools import partial
# med27_murinize = partial(murinize, entry=med27_chain)



from typing import (Tuple)

import pandas as pd

from ..weights import term_meanings


def get_lowest_contributor(row: pd.Series) -> Tuple[str, float]:
    """
    Not smallest/infinitesimal in abs contribution, but lowest number (i.e. most negative)
    """
    delta_names = [col for col in row.index if 'delta' in col]
    srow = row[delta_names].sort_values(ascending=True)
    return srow.index[0], srow[0]


def get_highest_contributor(row: pd.Series) -> Tuple[str, float]:
    """
    Not largest in abs contribution, but highest number (i.e. most positive)
    """
    delta_names = [col for col in row.index if 'delta' in col]
    srow = row[delta_names].sort_values(ascending=False)
    return srow.index[0], srow[0]


def get_largest_contributor(row: pd.Series) -> Tuple[str, float]:
    """
    Largest in abs amount as per the confusing fact that a very low negative number is large.
    """
    delta_names = [col for col in row.index if 'delta' in col]
    srow = row[delta_names].abs().sort_values(ascending=False)
    return srow.index[0], srow[0]


def extend_scores(scores: pd.DataFrame):
    """
    Adds the following fields:

    * highest/lowest_contributor
    * highest/lowest_contributor_value
    * highest/lowest_contributor_wordy

    :param scores: pd.DataFrame(output_of_variants)
    :return:
    """
    scores['highest_contributor'] = scores.apply(lambda row: get_highest_contributor(row)[0].replace('delta_', ''), 1)
    scores['highest_contributor_value'] = scores.apply(lambda row: row['delta_' + row.highest_contributor], 1)
    scores['highest_contributor_wordy'] = scores.apply(lambda row: term_meanings[row.highest_contributor], 1)

    scores['lowest_contributor'] = scores.apply(lambda row: get_lowest_contributor(row)[0].replace('delta_', ''), 1)
    scores['lowest_contributor_value'] = scores.apply(lambda row: row['delta_' + row.lowest_contributor], 1)
    scores['lowest_contributor_wordy'] = scores.apply(lambda row: term_meanings[row.lowest_contributor], 1)



import pyrosetta


class Remodel:

    find_neighbors = property(
        lambda self: pyrosetta.rosetta.basic.options.get_boolean_option('remodel:design:find_neighbors'),
        lambda self, value: pyrosetta.rosetta.basic.options.set_boolean_option('remodel:design:find_neighbors', value))


    generic_aa = property(
        lambda self: pyrosetta.rosetta.basic.options.get_string_option('remodel:generic_aa'),
        lambda self, value: pyrosetta.rosetta.basic.options.set_string_option('remodel:generic_aa', value))

    quick_and_dirty = property(
        lambda self: pyrosetta.rosetta.basic.options.get_boolean_option('quick_and_dirty'),
        lambda self, value: pyrosetta.rosetta.basic.options.set_boolean_option('remodel:quick_and_dirty', value))

    blueprint = property(
        lambda self: pyrosetta.rosetta.basic.options.set_file_option('remodel:blueprint'),
        lambda self, filename: pyrosetta.rosetta.basic.options.set_file_option('remodel:blueprint', filename))

    def get_remodelmover(self,
                         dr_cycles:int=3,
                         max_linear_chainbreak:float=0.07,
                         redesign_loop_neighborhood:bool=False):
        rm = pyrosetta.rosetta.protocols.forge.remodel.RemodelMover()
        rm.register_options()
        rm.dr_cycles(dr_cycles)
        rm.max_linear_chainbreak(max_linear_chainbreak)
        rm.redesign_loop_neighborhood(redesign_loop_neighborhood)
        return rm




import io
import logging
import pyrosetta
import re
from typing import (Any, Optional, Union, Dict, List)


def configure_logger() -> logging.Logger:
    """
    The function `get_logger`, simply adds a stringIO handler to the log and captures the log,
    thus making it easier to use.
    The function `get_log_entries`, spits out entries of a given level.

    :return: logger
    """
    pyrosetta.logging_support.set_logging_sink()
    logger = logging.getLogger("rosetta")
    logger.setLevel(logging.INFO) # default = logging.WARNING
    stringio = io.StringIO()
    handler = logging.StreamHandler(stringio)
    handler.setLevel(logging.INFO)
    handler.set_name('stringio')
    handler.setFormatter(logging.Formatter('[%(asctime)s] %(levelname)s - %(message)s'))
    logger.addHandler(handler)
    return logger


def get_log_entries(levelname: Union[str, int]=logging.INFO, query: Optional[str]=None) -> List[Dict[str, Any]]:
    """
    Get a list of all entries in log at a given level.
    levelname can be either an int (``logging.INFO`` etc. are numbers multiples of 10 in increasing severity)
    or a string of the level.
    Note that it is very crude: if INFO is requested, ERROR is not shown!

    :param levelname: int for the level number or str of the name
    :return: List of str
    """
    if isinstance(levelname, int):
        # logging.INFO is actually an int, not an enum
        levelnumber = logging.getLevelName(levelname)
        entries = get_all_log_entries()
        if query is None:
            return [e for e in entries if e['level'] == levelnumber]
        else:
            return [e for e in entries if str(query) in e['text']]

def get_all_log_entries() -> List[Dict[str, Any]]:
    stringio = logging.getLogger("rosetta").handlers[0].stream
    cleaned = []
    previous = None
    for entry in stringio.getvalue().split('\n'):
        rex = re.match(f'\[(.*)\] (.*) - (.*)', entry)
        if rex:
            if previous:
                cleaned.append(previous)
            previous = dict(datetime=rex.group(1),
                            level=logging.getLevelName(rex.group(2)),
                            text=rex.group(3))
        else:
            previous['text'] += '\n' + entry
    if previous:
        cleaned.append(previous)
    return cleaned


from .hbonds import get_hbond_dicts, hbond2dict, BondDataType
from .ss import get_ss
from .betaturn import get_betaturns
from .cis import get_cis_residues
from .xlinks import get_xlink_details, get_xlink_idx, is_xlinked




# This has several variants...

# this is a mutation, which is validated against a pose (G10W will fail if resi 10 is not glycine).
from .mutation import Mutation

# this generates and scores variant poses
# called Variant in some older repos
from .variant import MutantScorer
from .scores import extend_scores



import os
from typing import (Optional, Union, Dict, List)

import pandas as pd
import pyrosetta

from .terms import term_meanings


class WeightWatcher:
    """
    A class to easily look at the weights for the terms of different scorefunctions.
    It has various functionalities.

    List available scorefunctions:

    .. code-block:: python
      ww = WeightWatcher()
      print( ww.possible_scorefxn_names ) # dynamic attribute

    Find a scorefunction that mentions a word:

    .. code-block:: python
      ww.find_metion('foo')

    Get the comment block of a scorefunction:

    .. code-block:: python
      print(ww.get_scorefxn_comments('ref2015'))

    Get a scorefunction by name by calling the appropriate options first.

    .. code-block:: python
      scorefxn = ww.get_scorefxn('beta_nov16')

    Get weights (including ref) for a scorefunction or a name of one

    .. code-block:: python
      weights = ww.get_weights('beta_nov16')

    Get a pandas table of the weights

    .. code-block:: python
      weight_table = ww.compare(['ref2015', 'beta_nov16'], different_only=True)

    """
    folder = os.path.join(os.path.split(pyrosetta.__file__)[0], 'database', 'scoring', 'weights')
    term_meanings = term_meanings

    # def __init__(self):
    #     pyrosetta.distributed.maybe_init(extra_options='-mute all')

    def get_weight(self,
                   scorefxn: pyrosetta.ScoreFunction,
                   score_type: pyrosetta.rosetta.core.scoring.ScoreType
                   ) -> float:
        return scorefxn.get_weight(getattr(pyrosetta.rosetta.core.scoring.ScoreType, score_type.name))

    def get_weights(self,
                    scorefxn: Union[str, pyrosetta.ScoreFunction]
                    ) -> Dict[str, float]:
        """
        Gets weights for scorefxn
        """
        if isinstance(scorefxn, str):
            # scorefxn = pyrosetta.create_score_function(scorefxn)
            # mod:
            scorefxn = self.get_scorefxn(scorefxn)
        return {'name': scorefxn.get_name(),
                **{score_type.name: self.get_weight(scorefxn, score_type) for score_type in
                   scorefxn.get_nonzero_weighted_scoretypes()},
                **self.get_ref_values_badly(scorefxn, prefix=True)}

    def get_ref_values_badly(self,
                             scorefxn: pyrosetta.ScoreFunction,
                             prefix: Optional[bool] = True
                             ) -> Dict[str, float]:
        """
        I could not find a direct way to get_method_weights.
        Here this horrid way.
        """
        # similarity order 'IVLFCMAGTSWYPHNDEQKR'
        # name1 alphabetical order 'ACDEFGHIKLMNPQRSTVWY'
        # name3 alphabetical order 'ARNDCQEGHILKMFPSTWYV'
        pose = pyrosetta.pose_from_sequence('ARNDCQEGHILKMFPSTWYV')
        scorefxn(pose)
        aa_ref = {}
        for res in range(1, pose.total_residue() + 1):
            name = pose.residue(res).name3()
            value = pose.energies().residue_total_energies(res)[pyrosetta.rosetta.core.scoring.ScoreType.ref]
            if prefix:
                aa_ref[f'ref_{name}'] = value
            else:
                aa_ref[name] = value
        return aa_ref

    def get_scorefxn(self, scorefxn_name: str) -> pyrosetta.ScoreFunction:
        """
        Gets the scorefxn with appropriate corrections.
        """
        if isinstance(scorefxn_name, pyrosetta.ScoreFunction):
            return scorefxn_name  # it's already a scorefxn (not a name)
        corrections = {'beta_july15': False,
                       'beta_nov16': False,
                       'gen_potential': False,
                       'restore_talaris_behavior': False
                       }
        if 'beta_july15' in scorefxn_name or 'beta_nov15' in scorefxn_name:
            # beta_july15 is ref2015
            corrections['beta_july15'] = True
        elif 'beta_nov16' in scorefxn_name:
            corrections['beta_nov16'] = True
        elif 'genpot' in scorefxn_name:
            corrections['gen_potential'] = True
            pyrosetta.rosetta.basic.options.set_boolean_option('corrections:beta_july15', True)
        elif 'talaris' in scorefxn_name:  # 2013 and 2014
            corrections['restore_talaris_behavior'] = True
        else:
            pass
        for corr, value in corrections.items():
            pyrosetta.rosetta.basic.options.set_boolean_option(f'corrections:{corr}', value)
        return pyrosetta.create_score_function(scorefxn_name)

    @property
    def possible_scorefxn_names(self) -> List[str]:
        """
        Returns the scorefxn names
        """
        return sorted([fn.replace('.wts', '') for fn in os.listdir(self.folder) if '.wts' == os.path.splitext(fn)[1]])

    def get_scorefxn_block(self, scorefxn_name: str) -> str:
        """
        Read the file given a scorefxn name
        """
        if '/' in scorefxn_name:
            filename = scorefxn_name
        elif '.wts' in scorefxn_name:
            filename = os.path.join(self.folder, f'{scorefxn_name}')
        else:
            filename = os.path.join(self.folder, f'{scorefxn_name}.wts')
        with open(filename, 'r') as fh:
            return fh.read()

    def get_scorefxn_comments(self, scorefxn_name: str) -> str:
        """
        Get only the comments
        """
        lines = self.get_scorefxn_block(scorefxn_name).split('\n')
        return '\n'.join([line for line in lines if '#' in line])

    def find_metion(self, word: str) -> List[str]:
        """
        Find all the scorefxn names whose files contain the string word (case insensitive).

        >>> find_metion('spades')

        returns ``['hydrate_score12']``
        """
        mentionants = []
        for scorefxn_name in self.possible_scorefxn_names:
            block = self.get_scorefxn_block(scorefxn_name)
            if word.lower() in block.lower():
                mentionants.append(scorefxn_name)
        return mentionants

    def compare(self, scorefxn_names: List[str], different_only: bool = False) -> pd.DataFrame:
        terms = pd.DataFrame(map(self.get_weights, scorefxn_names)).fillna(0).set_index('name')
        if different_only:
            unique_cols = [col for col in terms.columns if len(terms[col].unique()) > 1]
            return terms[unique_cols]
        else:
            return terms


from __future__ import annotations

import pyrosetta
import re, os, csv, json
from typing import Optional, List, Dict, Union, Any, Callable, Tuple
from .mutation import Mutation


class MutantScorer:
    """
    Copy pasted from PI4KA <- GNB2 <- SnoopCatcher
    """

    # ===== init ===========================================================================

    def __init__(self,
                 pose: pyrosetta.Pose,
                 modelname: str,
                 scorefxn: Optional[pyrosetta.ScoreFunction] = None,
                 strict_about_starting_residue: bool = True,
                 verbose: bool = False):
        self.pose = pose
        self.modelname = modelname
        self.strict_about_starting_residue = bool(strict_about_starting_residue)
        if scorefxn is None:
            self.scorefxn = pyrosetta.get_fa_scorefxn()
        else:
            self.scorefxn = scorefxn
        self.verbose = verbose
        self.output_folder = 'variants'

    @classmethod
    def from_file(cls, filename: str, params_filenames: Optional[List[str]] = None, **kwargs):
        return cls(pose=cls._load_pose_from_file(filename, params_filenames), **kwargs)

    @classmethod
    def _load_pose_from_file(cls, filename: str, params_filenames: Optional[List[str]] = None) -> pyrosetta.Pose:
        """
        Loads a pose from filename with the params in the params_folder
        :param filename:
        :param params_filenames:
        :return:
        """
        pose = pyrosetta.Pose()
        if params_filenames:
            params_paths = pyrosetta.rosetta.utility.vector1_string()
            params_paths.extend(params_filenames)
            pyrosetta.generate_nonstandard_residue_set(pose, params_paths)
        pyrosetta.rosetta.core.import_pose.pose_from_file(pose, filename)
        return pose

    # ===== score =========================================================================
    def score_mutation(self,
                       mutation_name: str,
                       chains: str,
                       distance: int,
                       cycles: int,
                       interfaces,
                       ref_interface_dG: Dict[str, float],
                       final_func: Optional[Callable] = None,
                       preminimise: bool = False,
                       movement: bool = False) -> Tuple[Dict[str, float], pyrosetta.Pose, pyrosetta.Pose]:
        """
        Scores the mutation ``mutation_name`` (str or Mutation instance)
        returning three objects: a dict of scores, the wt (may differ from pose if preminimise=True) and mutant pose

        :param mutation_name:
        :param chains:
        :param distance:
        :param cycles:
        :param interfaces:
        :param ref_interface_dG: premade if no preminimise.
        :param final_func:
        :param preminimise:
        :return:
        """
        if preminimise:
            premutant = self.pose.clone()
        else:
            premutant = self.pose
        assert len(chains) > 0, 'Specify at least one chain e.g. "A"'
        for chain in chains:
            mutation = self.parse_mutation(mutation_name, chain)
            if self.verbose:
                print(mutation)
            if not self.does_contain(mutation):
                raise ValueError('Absent')
            if preminimise:
                self.relax_around_mover(premutant,
                                        mutation=mutation,
                                        distance=distance,
                                        cycles=cycles)
                if self.verbose:
                    print('preminimisation complete')
        variant = premutant.clone()
        for chain in chains:
            mutation = self.parse_mutation(mutation_name, chain)
            variant = self.make_mutant(variant,
                                       mutation=mutation,
                                       distance=distance,
                                       cycles=cycles,
                                       inplace=True)
        if self.verbose:
            print('mutant made')
        variant.dump_scored_pdb(f'{self.output_folder}/{self.modelname}.{mutation}.pdb', self.scorefxn)
        # score proper
        data = self.score_only(variant=variant,
                               reference=premutant,
                               mutation=mutation,
                               chains=chains,
                               distance=distance,
                               interfaces=interfaces,
                               ref_interface_dG=ref_interface_dG if not preminimise else dict(),
                               final_func=final_func,
                               movement=movement)
        # if ref_interface_dG (above) is empty score_only calcutates it, so it makes no diff why its empty.
        # ie. the ref_interface_dG variable here is empty or an empty argument is passed.
        return data, premutant, variant

    def score_only(self,
                   variant: pyrosetta.Pose,
                   reference: pyrosetta.Pose,
                   mutation: Mutation,
                   chains: str,
                   distance: int,
                   interfaces: List[Tuple[str, str]],
                   ref_interface_dG: Dict,
                   final_func: Optional[Callable] = None,
                   movement: bool=False) -> dict:
        unweighted_scorefxn = self.get_unweighted_scorefxn()
        wt_score = unweighted_scorefxn(reference)
        mut_score = unweighted_scorefxn(variant)
        neigh_vector = self.get_neighbour_vector(pose=reference,
                                                 resi=mutation.pose_resi,
                                                 chain=None,  # if chain is present, it assumes resi is pdb number.
                                                 distance=distance)
        neigh_wt_score = unweighted_scorefxn.get_sub_score(reference, neigh_vector)
        neigh_mut_score = unweighted_scorefxn.get_sub_score(variant, neigh_vector)
        data = {'model': self.modelname,
                'mutation': str(mutation),
                'complex_ddG': mut_score - wt_score,
                'complex_native_dG': wt_score,
                'complex_mutant_dG': mut_score,
                'complex_ddG_neigh_only': neigh_mut_score - neigh_wt_score,
                'complex_native_dG_only': neigh_wt_score,
                'complex_mutant_dG_only': neigh_mut_score,
                'FA_RMSD': self.FA_RMSD(self.pose,
                                        variant,
                                        resi=mutation.pose_resi,
                                        chain=None,   # None becase pose_resi is provided.
                                        distance=distance),
                'CA_RMSD': self.CA_RMSD(self.pose,
                                        variant,
                                        resi=mutation.pose_resi,
                                        chain=None,  # None because pose_resi is provided.
                                        distance=distance)
                }
        # interfaces
        for interface_name, interface_scheme in interfaces:
            if self.has_interface(variant, interface_scheme):
                if interface_name not in ref_interface_dG:
                    ref_interface_dG[interface_name] = self.score_interface(reference,
                                                                            interface_scheme)['interface_dG']
                if self.verbose:
                    print(f'{interface_name} ({interface_scheme}) applicable to {self.modelname}')
                i = self.score_interface(variant, interface_scheme)['interface_dG']
            else:
                print(f'{interface_name} ({interface_scheme}) not applicable to {self.modelname}')
                i = float('nan')
            data[f'{interface_name}_interface_native_dG'] = ref_interface_dG[interface_name]
            data[f'{interface_name}_interface_mutant_dG'] = i
            data[f'{interface_name}_interface_ddG'] = i - ref_interface_dG[interface_name]
        if self.verbose:
            print('interface scored')
        # raw
        wt_scoredex = self.get_wscoredict(reference)
        mut_scoredex = self.get_wscoredict(variant)
        delta_scoredex = self.delta_scoredict(mut_scoredex, wt_scoredex)
        if self.verbose:
            print('scores stored')
        # movement
        if movement:
            data['wt_rmsd'] = self.movement(original=reference, resi=mutation.pdb_resi, chain=chains[0], distance=distance)
            data['mut_rmsd'] = self.movement(original=reference, resi=mutation.pdb_resi, chain=chains[0], distance=distance)
            data['ratio_rmsd'] = data['mut_rmsd'] / data['wt_rmsd']
            if self.verbose:
                print('movement assessed')
        data = {**data,
                **self.prefix_dict(wt_scoredex, 'wt'),
                **self.prefix_dict(mut_scoredex, 'mut'),
                **self.prefix_dict(delta_scoredex, 'delta')}
        if final_func is not None:  # callable
            final_func(data, reference, variant)
            if self.verbose:
                print('extra step done.')
        return data

    def make_output_folder(self):
        if not os.path.exists(self.output_folder):
            os.mkdir(self.output_folder)

    def get_unweighted_scorefxn(self):
        unweighted_scorefxn = self.scorefxn.clone()
        ST = pyrosetta.rosetta.core.scoring.ScoreType
        unweighted_scorefxn.set_weight(ST.atom_pair_constraint, 0)
        unweighted_scorefxn.set_weight(ST.angle_constraint, 0)
        unweighted_scorefxn.set_weight(ST.coordinate_constraint, 0)
        unweighted_scorefxn.set_weight(ST.dihedral_constraint, 0)
        return unweighted_scorefxn


    def score_mutations(self,
                        mutations,
                        chains='A',
                        interfaces=(),  # list of two: name, scheme
                        preminimise=False,
                        distance=10,
                        cycles=5,
                        final_func: Optional[Callable] = None) -> List[Dict[str, Union[float, str]]]:
        self.make_output_folder()
        data = []
        ## wt
        ref_interface_dG = {}
        scores = {}  # not written to csv file.
        if not preminimise:
            # touch the energies:
            self.get_unweighted_scorefxn()(self.pose)
            for interface_name, interface_scheme in interfaces:
                ref_interface_dG[interface_name] = self.score_interface(self.pose, interface_scheme)['interface_dG']
        else:
            pass  # calculated for each.
        ## muts
        for mutation_name in mutations:
            try:
                datum, premutant, mutant = self.score_mutation(mutation_name=mutation_name,
                                                               chains=chains,
                                                               distance=distance,
                                                               cycles=cycles,
                                                               interfaces=interfaces,
                                                               ref_interface_dG=ref_interface_dG,
                                                               final_func=final_func,
                                                               preminimise=preminimise)
            except Exception as error:
                msg = f"{error.__class__.__name__}: {error}"
                print(msg)
                datum = {'model': self.modelname,
                         'mutation': str(mutation_name),
                         'complex_ddG': msg
                         }
            data.append(datum)
        return data

    # ===== mutation ======================================================================

    def parse_mutation(self, mutation: Union[str, Mutation], chain, pose: pyrosetta.Pose = None):
        if pose is None:
            pose = self.pose
        if isinstance(mutation, str):
            mutant = Mutation(mutation, chain, pose)
        elif isinstance(mutation, Mutation):
            mutant = mutation
        else:
            raise TypeError(f'Does not accept mutation of type {mutation.__class__.__name__}')
        if mutant.pose_resi == 0:
            raise ValueError('Not in pose')
        if self.strict_about_starting_residue:
            mutant.assert_valid()
        return mutant

    def make_mutant(self,
                    pose: pyrosetta.Pose,
                    mutation: Union[str, Mutation],
                    chain='A',
                    distance: int = 10,
                    cycles: int = 5,
                    inplace: bool = False,
                    ) -> pyrosetta.Pose:
        """
        Make a point mutant (``A23D``).

        :param pose: pose
        :param mutation:
        :param chain:
        :param inplace:
        :return: a copy if inplace is false
        """
        if pose is None and inplace:
            raise ValueError('inplace parameter = True works only if pose is provide')
        elif pose is None:
            mutant = self.pose.clone()
        elif not inplace:
            mutant = pose.clone()
        else:
            mutant = pose
        if isinstance(mutation, str):
            mutation = Mutation(mutation, chain, mutant)
        MutateResidue = pyrosetta.rosetta.protocols.simple_moves.MutateResidue
        MutateResidue(target=mutation.pose_resi, new_res=mutation.to_resn3).apply(mutant)
        self.relax_around_mover(mutant,
                                mutation=mutation,
                                distance=distance,
                                cycles=cycles,
                                own_chain_only=False)
        return mutant

    def does_contain(self, mutation: Union[Mutation, str], chain: Optional[str] = None) -> bool:
        assert isinstance(mutation, Mutation) or chain is not None, 'mutation as str requires chain.'
        if isinstance(mutation, str):
            mutation = Mutation(mutation, chain, self.pose)
        if mutation.pose_resi == 0:
            return False
        if mutation.pose_resn1 not in mutation._name3.keys():
            return False
        else:
            return True

    def relax_around_mover(self,
                           pose: pyrosetta.Pose,
                           mutation: Optional[Mutation] = None,
                           resi: int = None, chain: str = None, cycles=5, distance=5,
                           own_chain_only=False) -> None:
        """
        Relaxes pose ``distance`` around resi:chain or mutation

        :param resi: PDB residue number.
        :param chain:
        :param pose:
        :param cycles: of relax (3 quick, 15 thorough)
        :param distance:
        :return:
        """
        if mutation is None and resi is None:
            raise ValueError('mutation or resi+chain required')
        elif mutation is not None:
            resi = mutation.pose_resi
            chain = None  # chain None is for pose resi
        else:
            pass
        if pose is None:
            pose = self.pose
        movemap = pyrosetta.MoveMap()
        ####
        n = self.get_neighbour_vector(pose=pose, resi=resi, chain=chain, distance=distance,
                                      own_chain_only=own_chain_only)
        # print(pyrosetta.rosetta.core.select.residue_selector.ResidueVector(n))
        movemap.set_bb(False)
        movemap.set_bb(allow_bb=n)
        movemap.set_chi(False)
        movemap.set_chi(allow_chi=n)
        movemap.set_jump(False)

        relax = pyrosetta.rosetta.protocols.relax.FastRelax(self.scorefxn, cycles)
        relax.set_movemap(movemap)
        relax.set_movemap_disables_packing_of_fixed_chi_positions(True)
        if self.scorefxn.get_weight(pyrosetta.rosetta.core.scoring.ScoreType.cart_bonded) > 0:
            # it's cartesian!
            relax.cartesian(True)
            relax.minimize_bond_angles(True)
            relax.minimize_bond_lengths(True)
        else:
            relax.cartesian(False)
        relax.apply(pose)

    def get_neighbour_vector(self, pose: pyrosetta.Pose, resi: int, chain: str, distance: int,
                             include_focus_in_subset: bool = True,
                             own_chain_only: bool = False) -> pyrosetta.rosetta.utility.vector1_bool:
        resi_sele = pyrosetta.rosetta.core.select.residue_selector.ResidueIndexSelector()
        if chain is None:  # pose numbering.
            resi_sele.set_index(resi)
        else:
            resi_sele.set_index(pose.pdb_info().pdb2pose(chain=chain, res=resi))
        NeighborhoodResidueSelector = pyrosetta.rosetta.core.select.residue_selector.NeighborhoodResidueSelector
        neigh_sele = NeighborhoodResidueSelector(resi_sele, distance=distance,
                                                 include_focus_in_subset=include_focus_in_subset)
        if own_chain_only and chain is not None:
            chain_sele = pyrosetta.rosetta.core.select.residue_selector.ChainSelector(chain)
            and_sele = pyrosetta.rosetta.core.select.residue_selector.AndResidueSelector(neigh_sele, chain_sele)
            return and_sele.apply(pose)
        else:
            return neigh_sele.apply(pose)

    # ===== interface ======================================================================

    def score_interface(self, pose: pyrosetta.Pose, interface: str) -> Dict[str, float]:
        if pose is None:
            pose = self.pose
        assert self.has_interface(pose, interface), f'There is no {interface}'
        ia = pyrosetta.rosetta.protocols.analysis.InterfaceAnalyzerMover(interface)
        ia.apply(pose)
        return {'complex_energy': ia.get_complex_energy(),
                'separated_interface_energy': ia.get_separated_interface_energy(),
                'complexed_sasa': ia.get_complexed_sasa(),
                'crossterm_interface_energy': ia.get_crossterm_interface_energy(),
                'interface_dG': ia.get_interface_dG(),
                'interface_delta_sasa': ia.get_interface_delta_sasa()}

    def has_interface(self, pose: pyrosetta.Pose, interface: str) -> bool:
        have_chains = self.get_present_chains(pose)
        want_chains = set(interface.replace('_', ''))
        return have_chains == want_chains

    def get_present_chains(self, pose: Optional[pyrosetta.Pose] = None):
        if pose is None:
            pose = self.pose
        # pose2pdb = pose.pdb_info().pose2pdb
        # return {pose2pdb(r).split()[1] for r in range(1, pose.total_residue() + 1)}
        pdb_info = pose.pdb_info()
        return {pdb_info.chain(res + 1) for res in range(pose.total_residue())}

    def has_residue(self, pose: pyrosetta.Pose, resi: int, chain: str) -> bool:
        if pose is None:
            pose = self.pose
        pdb2pose = pose.pdb_info().pdb2pose
        r = pdb2pose(res=resi, chain=chain)
        return r != 0

    # ========= RMSD relative to native ===========================================

    def vector2list(self, vector: pyrosetta.rosetta.utility.vector1_bool) -> pyrosetta.rosetta.std.list_unsigned_long_t:
        rv = pyrosetta.rosetta.core.select.residue_selector.ResidueVector(vector)
        x = pyrosetta.rosetta.std.list_unsigned_long_t()
        assert len(rv) > 0, 'Vector is empty!'
        for w in rv:
            x.append(w)
        return x

    def CA_RMSD(self, poseA: pyrosetta.Pose, poseB: pyrosetta.Pose, resi: int, chain: Union[str, None], distance: int) -> float:
        n = self.get_neighbour_vector(pose=poseA, resi=resi, chain=chain, distance=distance)
        residues = self.vector2list(n)
        return pyrosetta.rosetta.core.scoring.CA_rmsd(poseA, poseB, residues)

    def FA_RMSD(self, poseA: pyrosetta.Pose, poseB: pyrosetta.Pose, resi: int, chain: Union[str, None], distance: int) -> float:
        n = self.get_neighbour_vector(pose=poseA, resi=resi, chain=chain, distance=distance,
                                      include_focus_in_subset=False)
        residues = self.vector2list(n)
        # pyrosetta.rosetta.core.scoring.automorphic_rmsd(residueA, residueB, False)
        return pyrosetta.rosetta.core.scoring.all_atom_rmsd(poseA, poseB, residues)

    # ======= score dictionary operations =================================================
    def get_scoredict(self, pose: pyrosetta.Pose) -> Dict[str, float]:
        """
        Given a pose get the global scores.
        """
        a = pose.energies().total_energies_array()
        return dict(zip(a.dtype.fields.keys(), a.tolist()[0]))

    def get_wscoredict(self, pose: pyrosetta.Pose) -> Dict[str, float]:
        scoredex = self.get_scoredict(pose)
        stm = pyrosetta.rosetta.core.scoring.ScoreTypeManager()
        return {k: scoredex[k] * self.scorefxn.get_weight(stm.score_type_from_name(k)) for k in scoredex}

    def delta_scoredict(self, minuend: Dict[str, float], subtrahend: Dict[str, float]) -> Dict[str, float]:
        """
        minuend - subtrahend = difference
        given two dict return the difference -without using pandas.
        """
        assert all([isinstance(v, float) for v in [*minuend.values(), *subtrahend.values()]]), 'Float only, please...'
        minuend_keys = set(minuend.keys())
        subtrahend_keys = set(subtrahend.keys())
        common_keys = minuend_keys & subtrahend_keys
        minuend_unique_keys = minuend_keys - subtrahend_keys
        subtrahend_unique_keys = minuend_keys - subtrahend_keys
        return {**{k: minuend[k] - subtrahend[k] for k in common_keys},  # common
                **{k: minuend[k] - 0 for k in minuend_unique_keys},  # unique
                **{k: 0 - subtrahend[k] for k in subtrahend_unique_keys}  # unique
                }

    def prefix_dict(self, dex: Dict[str, Any], prefix: str) -> Dict[str, Any]:
        return {f'{prefix}_{k}': v for k, v in dex.items()}

    # ====== movement ==============================================================================

    def movement(self, original: pyrosetta.Pose, resi: int, chain: str, distance: int,
                 trials: int = 50, temperature: int = 1.0, replicate_number: int = 10):
        """
        This method adapted from a notebook of mine, but not from an official source, is not well written.
        It should be a filter and score combo.

        It returns the largest bb_rmsd of the pdb residue resi following backrub.
        """
        # this code is experimental

        n = self.get_neighbour_vector(pose=original, resi=resi, chain=chain, distance=distance,
                                      own_chain_only=False)
        # resi
        if chain is None:  # pose numbering.
            target_res = resi
        else:
            target_res = original.pdb_info().pdb2pose(chain=chain, res=resi)
        # prep
        rv = pyrosetta.rosetta.core.select.residue_selector.ResidueVector(n)
        backrub = pyrosetta.rosetta.protocols.backrub.BackrubMover()
        backrub.set_pivot_residues(rv)
        # https://www.rosettacommons.org/docs/latest/scripting_documentation/RosettaScripts/Movers/movers_pages/GenericMonteCarloMover
        monégasque = pyrosetta.rosetta.protocols.monte_carlo.GenericMonteCarloMover(maxtrials=trials,
                                                                                    max_accepted_trials=trials,
                                                                                    # gen.max_accepted_trials() = 0
                                                                                    task_scaling=5,
                                                                                    # gen.task_scaling()
                                                                                    mover=backrub,
                                                                                    temperature=temperature,
                                                                                    sample_type='low',
                                                                                    drift=True)
        monégasque.set_scorefxn(self.scorefxn)
        # monégasque.add_filter(filters , False , 0.005 , 'low'  , True )
        # define the first 4 atoms (N C CA O)
        am = pyrosetta.rosetta.utility.vector1_unsigned_long(4)
        for i in range(1, 5):
            am[i] = i
        # find most deviant
        best_r = 0
        for i in range(replicate_number):
            variant = original.clone()
            monégasque.apply(variant)
            if monégasque.accept_counter() > 0:
                variant = monégasque.last_accepted_pose()  # pretty sure redundant
                # bb_rmsd is all residues: pyrosetta.rosetta.core.scoring.bb_rmsd(pose, ori)
                r = pyrosetta.rosetta.core.scoring.residue_rmsd_nosuper(variant.residue(target_res),
                                                                        original.residue(target_res), am)
                if r > best_r:
                    best_r = r
        return best_r



from functools import partial
from typing import (Callable, List, Dict)
import pyrosetta

from .ss import get_ss
from ..common_ops import pose_range

def get_betaturns(pose: pyrosetta.Pose) -> Dict[int, str]:
    """
    There is a wee problem in that alpha-helices get classified as beta-turns
    And I am not sure all are beta-turns...
    """

    get_ss(pose)  # fill secstruct
    btd = pyrosetta.rosetta.protocols.features.BetaTurnDetection()
    beta_turn_present: Callable[[int], bool] = partial(btd.beta_turn_present, pose)
    loop_iter = filter(lambda i: pose.secstruct(i) == 'L', pose_range(pose, protein_only=True))
    maxed = list(loop_iter)[:-3]
    beta_iter = filter(beta_turn_present, maxed)
    return {i: btd.beta_turn_type(pose, i) for i in beta_iter}


from typing import Union, List
import pyrosetta


class BlueprinterSubscripted:
    def __iter__(self):
        return filter(len, self.rows)

    def __getitem__(self, idx: Union[int, slice]) -> list:
        if isinstance(idx, int):
            for row in self:
                i = int(row[0])
                if i == idx:
                    return row
        elif isinstance(idx, slice):
            sliced = []
            for row in self:
                i = int(row[0])
                if i >= idx.start and i <= idx.stop:
                    sliced.append(row)
            return sliced
        else:
            raise TypeError(f'Expected int or slice, got a {type(idx)}')
        raise ValueError('index not found.')

    def _setrow(self, row, value):
        idx = int(row[0])
        del row[2:]
        row.append(self.get_ss(idx))
        row.append(str(value))
        return None

    def __setitem__(self, idx: Union[int, slice], value: str) -> None:
        if isinstance(idx, int):
            row = self[idx]
            self._setrow(row, value)
        elif isinstance(idx, slice):
            sliced = []
            for row in self[idx]:
                if '*' in value:
                    # star is a wildcard to mean same as original
                    self._setrow(row, value.replace('*', row[1]))
                else:
                    self._setrow(row, value)
        else:
            raise TypeError(f'Expected int or slice, got a {type(idx)}')

    def __delitem__(self, idx: Union[int, slice]) -> list:
        """
        It is not quite a deletion as the 'row' is still there but empty

        :param idx:
        :return:
        """
        deleted = []
        if isinstance(idx, int):
            row = self[idx]
            deleted.append(row.copy())
            del row[:]
        elif isinstance(idx, slice):
            for row in self[idx]:
                deleted.append(row.copy())
                del row[:]
        else:
            raise TypeError(f'Expected int or slice, got a {type(idx)}')
        return deleted

    @property
    def max(self):
        """
        :return: Highest pose index (this is different from the number of lines due to indels)
        """
        return max([int(row[0]) for row in self])

    def insert(self, idx: int, value: Union[str, List[str]], before: bool = True):
        """

        :param idx:
        :param value:
        :param before: If true, like the list method ``insert`` it inserts before
        :return:
        """
        # homogenise inputs
        if isinstance(value, str):
            values = [value]
        elif isinstance(value, list):
            values = value
        else:
            raise TypeError(f'{type(value)} is not Union[str, List[str]]')
        # insert
        for i, row in enumerate(self):
            if idx == int(row[0]):
                for j, val in enumerate(values):
                    inserendum = [0, 'X', 'D', val]
                    if before:
                        self.rows.insert(i + j, inserendum)
                    else:
                        self.rows.insert(i + j + 1, inserendum)
                break
        else:
            raise ValueError(f'Index {idx} not found.')
        # wobble sides
        if before:
            other_idx = idx - 1
        elif self.max >= idx + 1:
            other_idx = idx + 1
        else:
            other_idx = idx  # does nothing.
        for i in (idx, other_idx):
            if len(self[i]) <= 3:
                self[i] = f'PIKAA {self[1]}'

    def insert_before(self, idx: int, value: Union[str, List[str]]):
        self.insert(idx, value, before=True)

    def insert_after(self, idx: int, value: Union[str, List[str]]):
        self.insert(idx, value, before=False)

    def prepend(self, value: Union[str, List[str]]):
        self.insert(1, value, before=True)

    def append(self, value: Union[str, List[str]]):
        self.insert(self.max, value, before=False)

    # ===== output ==============================

    def __str__(self):
        return '\n'.join([' '.join(map(str, row)) for row in self])

    def to_pdb_str(self, pose):
        pose2pdb = pose.pdb_info().pose2pdb
        lines = []
        for row in self:
            if row[0] == 0:
                lines.append(' '.join(map(str, row)))
            else:
                resi, chain = pose2pdb(row[0]).split()
                lines.append(f'{resi}{chain} '+' '.join(row[1:]))
        return '\n'.join(lines)

    def write(self, bluprint_filename: str):
        with open(bluprint_filename, 'w') as w:
            w.write(str(self))

    def write_pdb_numbered(self, bluprint_filename: str, pose: pyrosetta.Pose):
        with open(bluprint_filename, 'w') as w:
            w.write(self.to_pdb_str(pose))


    def set(self, bluprint_filename: str = 'model.blu'):
        """
        Writes and Sets the blueprint in options
        """
        self.write(bluprint_filename)
        self.blueprint = bluprint_filename



#

from typing import (Optional, Tuple, Dict)

import pyrosetta
import warnings


class AtomicInteractions:
    """
    Gets the per atom energies for the per_atom.

    >>> ai = AtomicInteractions(pose, 1)  # noqa  - pose residue index 1
    >>> print(ai.describe_best())
    >>> ai.per_atom[(' N  ', 2, ' N  ')]

    Unfortunately, bonding is not taken into account therefore the total is more favourable
    as these are ignored.

    >>> ai.total, ai.expected_total
    """

    score_types = ['fa_atr', 'fa_rep', 'fa_sol', 'fa_elec']
    term_relevance_cutoff = 0.5

    def __init__(self,
                 pose: pyrosetta.Pose,
                 target_idx: int,
                 threshold: int = 3,
                 scorefxn: Optional[pyrosetta.ScoreFunction] = None,
                 weighted: bool = True,
                 halved: bool = False):
        # --- input
        self.pose = pose
        self.target_idx = target_idx
        self.target_residue = self.pose.residue(self.target_idx)
        assert self.target_residue, f'{self.target_idx} not in pose'
        self.threshold = threshold
        if scorefxn is None:
            self.scorefxn = pyrosetta.get_fa_scorefxn() # noqa
        else:
            self.scorefxn = scorefxn
        self.weighted = weighted
        stm = pyrosetta.rosetta.core.scoring.ScoreTypeManager()
        get_weights = lambda name: self.scorefxn.get_weight(stm.score_type_from_name(name))
        if not weighted:
            self.weights = {name: 1. for name in self.score_types}
        else:
            self.weights = {name: get_weights(st_name) for name, st_name in
                            zip(self.score_types, ['fa_atr', 'fa_rep', 'fa_sol', 'fa_elec'])}
        self.halved = halved
        # --- derived
        self.neighbours = self._get_neighbours()  # ResidueVector
        self.interactions = self._get_interactions()  # Dict[Tuple(str, int, str), Dict[str, float]]

    def _get_interactions(self) -> Dict[Tuple[str, int, str], Dict[str, float]]:
        interactions = {}
        ratio = 2 if self.halved else 1
        # Iterate per target residue's atom per all other residues' atoms
        for i in range(1, self.target_residue.natoms() + 1):
            if self.target_residue.atom_type(i).element() == 'X':
                continue
            iname = self.target_residue.atom_name(i)
            interactions[iname] = {}
            for r in self.neighbours: # noqa
                if r == self.target_idx:
                    continue
                other = self.pose.residue(r)
                interactions[iname][r] = {}
                for o in range(1, other.natoms() + 1):
                    oname = other.atom_name(o)
                    if r == self.target_idx and o == i:
                        continue  # self to self should be zero...!
                    elif other.atom_type(o).element() == 'X':
                        continue
                    score = pyrosetta.toolbox.atom_pair_energy.etable_atom_pair_energies(self.target_residue,
                                                                                         i,
                                                                                         other,
                                                                                         o,
                                                                                         self.scorefxn)
                    # per_atom[iname][r][oname] = dict(zip(score_types, score))
                    interactions[iname][r][oname] = {st: s * self.weights[st] / ratio
                                                     for st, s in zip(self.score_types, score)}
        # correct fa_rep for bonded
        for atomname, other_residue_index, other_atomname in self._get_connections():
            # per_atom[atomname][other_residue_index][other_atomname]['fa_rep'] = float('nan')
            del interactions[atomname][other_residue_index][other_atomname]
        # reshape
        reshaped = {(target_atomname, other_resi, atomname): interactions[target_atomname][other_resi][atomname] for
                    target_atomname in interactions for other_resi in interactions[target_atomname] for atomname in
                    interactions[target_atomname][other_resi]}
        return reshaped

    def _get_neighbours(self) -> pyrosetta.rosetta.core.select.residue_selector.ResidueVector:
        cc_sele = self.get_cc_selector()
        neighs = pyrosetta.rosetta.core.select.residue_selector.ResidueVector(cc_sele.apply(self.pose))
        return neighs

    def get_cc_selector(self):
        resi_sele = self.get_target_selector()
        cc_sele = pyrosetta.rosetta.core.select.residue_selector.CloseContactResidueSelector()
        cc_sele.central_residue_group_selector(resi_sele)
        cc_sele.threshold(self.threshold)
        return cc_sele

    def get_target_selector(self):
        return pyrosetta.rosetta.core.select.residue_selector.ResidueIndexSelector(self.target_idx)

    def _get_connections(self):
        residue = self.pose.residue(self.target_idx)
        for conn_id in range(1, residue.n_current_residue_connections() + 1):
            atomno = residue.residue_connect_atom_index(conn_id)
            atomname = residue.atom_name(atomno)
            adjecent_atomnos = residue.bonded_neighbor(atomno)
            other_residue_index = residue.residue_connection_partner(conn_id)
            other_residue = self.pose.residue(other_residue_index)
            other_atomname, other_atomno = self._get_other_connecting_atom(conn_id)
            other_adjecent_atomnos = other_residue.bonded_neighbor(other_atomno)
            return [(atomname, other_residue_index, other_atomname), ] + \
                   [(residue.atom_name(no), other_residue_index, other_atomname) for no in adjecent_atomnos] + \
                   [(atomname, other_residue_index, other_residue.atom_name(no)) for no in other_adjecent_atomnos]

    def _get_other_connecting_atom(self, conn_id) -> Tuple[str, int]:
        other_residue_index = self.target_residue.residue_connection_partner(conn_id)
        other_residue = self.pose.residue(other_residue_index)
        for other_conn_id in range(1, other_residue.n_current_residue_connections() + 1):
            if other_residue.residue_connection_partner(other_conn_id) == self.target_idx:
                other_atom_index = other_residue.residue_connect_atom_index(other_conn_id)
                return other_residue.atom_name(other_atom_index), other_atom_index
        else:
            raise ValueError('No connection found?!')


    @property
    def best_interactions(self):
        # dict is ordered...
        get_max = lambda k: max(map(abs, self.interactions[k].values()))
        return {k: self.interactions[k] for k in sorted(self.interactions, key=get_max, reverse=True)
                if get_max(k) >= self.term_relevance_cutoff}

    def describe_atom(self, residue, atomname):
        atomno = residue.atom_index(atomname)
        atomtype = residue.atom_type(atomno)
        verdict = {'heavyatom': atomtype.is_heavyatom(),
                   'polar hydrogen': atomtype.is_polar_hydrogen(),
                   'H-acceptor': atomtype.is_acceptor(),
                   'H-donor': atomtype.is_donor(),
                   'aromatic': atomtype.is_aromatic()}
        return ', '.join([atomtype.atom_type_name(), atomtype.element()] +
                         [k for k, v in verdict.items() if v])

    def describe_interaction(self, target_atomname, other_resi, other_atomname):
        other_residue = self.pose.residue(other_resi)
        target_text = self.describe_atom(self.target_residue, target_atomname)
        other_text = self.describe_atom(other_residue, other_atomname)
        # [('fa_sol', 0.7810698032540209), ('fa_elec', 0.37053877404394236)]
        all_terms = self.interactions[target_atomname, other_resi, other_atomname]
        # sort & filter | > 0.25 |
        terms = sorted(filter(lambda t: abs(t[1]) >= self.term_relevance_cutoff,
                              all_terms.items(),
                              ),
                       key=lambda t: -abs(t[1]))
        term_text = ' + '.join([f'{score_type} ({value:.1f} kcal/mol)' for score_type, value in terms])
        return f'{self.target_idx}.{target_atomname.strip()} ({target_text}) - ' + \
               f'{other_resi}.{other_atomname.strip()} ({other_text}) ' + \
               term_text

    def describe_best(self):
        return '\n'.join([self.describe_interaction(*k) for k in self.best_interactions])

    @property
    def expected_total(self):
        self.scorefxn(self.pose)
        return self.scorefxn.get_sub_score(self.pose, self.get_target_selector().apply(self.pose))

    @property
    def total(self):
        return sum([sum(d.values()) for d in self.interactions.values()])

class NeighbourInteractions:
    # old name!

    def __init__(self, *args, **kwargs):
        warnings.warn(f'`NeighbourInteractions` renamed to `AtomicInteractions`', DeprecationWarning)
        super().__init__(*args, **kwargs)








__all__ = ['superimpose_by_pLDDT']

# ============= pLDDT superimposition ======================================
# modified from https://gist.github.com/asford/c2404c8b045700f016fda8893325c807
import difflib
import pyrosetta
from pyrosetta.rosetta.std import map_core_id_AtomID_core_id_AtomID
from pyrosetta.rosetta.core.id import AtomID


def paired_residue_inds(a: pyrosetta.Pose, b: pyrosetta.Pose):
    """Get paired indicies of common residues in two structures."""
    aseq = a.sequence()
    bseq = b.sequence()
    astart, bstart, align_len = difflib.SequenceMatcher(a=aseq, b=bseq).find_longest_match(0, len(aseq), 0, len(bseq))
    return [(astart + i, bstart + i) for i in range(align_len)]


def superimpose_by_pLDDT(pose: pyrosetta.Pose,
                         original: pyrosetta.Pose,
                         cutoff=70,
                         pose_range=None) -> map_core_id_AtomID_core_id_AtomID:
    """
    Superimpose two poses, based on residues with pLDDT above a given threshold.

    :param pose:
    :param original:
    :param cutoff: %
    :param pose_range: optional argument to subset (start:int, end:int)
    :return:
    """
    ca_map = map_core_id_AtomID_core_id_AtomID()
    for w, a in paired_residue_inds(pose, original):
        if original.pdb_info().bfactor(a + 1, 1) <= cutoff:
            continue
        if pose_range is not None and (w < pose_range[0] or w > pose_range[1]):
            continue
        ca_map[AtomID(pose.residue(w + 1).atom_index("CA"), w + 1)] = AtomID(
            original.residue(a + 1).atom_index("CA"), a + 1
        )
    assert len(ca_map), 'No atoms greater than cutoff'
    pyrosetta.rosetta.core.scoring.superimpose_pose(pose, original, ca_map)
    return ca_map



__all__ = ['pose_from_alphafold2', 'get_alphafold2_error', 'reshape_errors']

from typing import (Union, Dict, List)

import numpy as np
import pyrosetta
import requests


def pose_from_alphafold2(uniprot: str) -> pyrosetta.Pose:
    reply = requests.get(f'https://alphafold.ebi.ac.uk/files/AF-{uniprot}-F1-model_v2.pdb')
    assert reply.status_code == 200
    pdbblock = reply.text
    pose = pyrosetta.Pose()
    pyrosetta.rosetta.core.import_pose.pose_from_pdbstring(pose, pdbblock)
    return pose


# ==== errors based methods ====================================================
def get_alphafold2_error(uniprot: str, reshaped=True) -> Union[np.ndarray, list]:
    """
    Returns the distances errors either as numpy matrix (``reshaped=True``)
    or as the weird format from AF2-EBI —see ``help(pyrosetta_help.alphafold.retrieval.reshape_errors)`` for more.

    Remember that the matrix is zero indexed and that these values are in Ångström
    and are not pLDDT, which are stored as b-factors.
    """
    # https://alphafold.ebi.ac.uk/files/AF-Q00341-F1-predicted_aligned_error_v1.json
    reply = requests.get(f'https://alphafold.ebi.ac.uk/files/AF-{uniprot}-F1-predicted_aligned_error_v2.json')
    assert reply.status_code == 200
    errors = reply.json()
    if reshaped:
        return reshape_errors(errors)
    else:
        return errors


def reshape_errors(errors: List[Dict[str, list]]) -> np.array:
    """
    The JSON from AF2 has a single element list.
    the sole element is a dictionary with keys 'residue1', 'residue2' and 'distance'.
    This method returns a matrix of distances reshaped based on the stated residue indices.
    This is rather unlikely to differ from a regular reshape...
    but idiotically I am not taking changes assuming it is always sorted.
    """
    n_residues = int(np.sqrt(len(errors[0]['distance'])))
    error_matrix = np.zeros((n_residues, n_residues)) * np.nan
    for i, d in enumerate(errors[0]['distance']):
        error_matrix[errors[0]['residue1'][i] - 1, errors[0]['residue2'][i] - 1] = d
    return error_matrix



import re
from typing import (Optional)

import pyrosetta

__all__ = ['get_NGL_selection_from_AtomID',
           'print_constraint_score',
           'print_constraint_scores',
           'get_AtomID',
           'get_AtomID_by_NGL_sele',
           'get_AtomID_from_pymol_line',
           'make_constraint_from_pymol_line',
           'print_bad_constraint_scores',
           'constraints2pandas']


def get_NGL_selection_from_AtomID(pose: pyrosetta.Pose, atom_id: pyrosetta.AtomID, named: bool = False):
    """
    Given a pyrosetta AtomID give an NGL selection.
    NB ``named`` gives the residue name (``'[SER]3:A.CA'``) but is not a valid selection.

    :param pose:
    :param atom_id:
    :param named:
    :return:
    """
    pose_resi = atom_id.rsd()
    residue = pose.residue(pose_resi)
    atom_name = residue.atom_name(atom_id.atomno()).strip()
    pdb_resi, chain = pose.pdb_info().pose2pdb(pose_resi).strip().split()
    if named:
        return f'[{residue.name3().strip()}]{pdb_resi}:{chain}.{atom_name}'
    else:
        return f'{pdb_resi}:{chain}.{atom_name}'


def print_constraint_score(pose: pyrosetta.Pose, con):
    """
    Print the constraint details. atoms and score

    :param pose:
    :param con:
    :return:
    """
    data = get_constraint_score_data(pose, con)
    print('{constraint_name}: {atom_A} – {atom_B} {function_name}={funpart} --> {score:.2}'.format(**data))


def get_constraint_score_data(pose: pyrosetta.Pose, con) -> dict:
    """
    Print the constraint details. atoms and score

    :param pose:
    :param con:
    :return:
    """
    a = get_NGL_selection_from_AtomID(pose, con.atom1())
    b = get_NGL_selection_from_AtomID(pose, con.atom2())
    fun = con.get_func()
    if hasattr(fun, 'x0') and hasattr(fun, 'sd'):
        funpart = f'{fun.x0():.2f}/{fun.sd():.2f}'
    elif hasattr(fun, 'x0'):
        funpart = f'{fun.x0():.2f}'
    else:
        funpart = f'NA'
    score = con.score(pose)
    return dict(constraint_name=con.__class__.__name__,
                function_name=fun.__class__.__name__,
                atom_A=a,  # NGL_selection
                atom_B=b,  # NGL_selection
                funpart=funpart,
                score=score
                )


def print_constraint_scores(pose: pyrosetta.Pose):
    """
    Prints the scores for each constraint in the pose

    :param pose:
    :return:
    """
    cs = pose.constraint_set()
    for con in cs.get_all_constraints():
        print_constraint_score(pose, con)


def constraints2pandas(pose):
    import pandas as pd
    cs = pose.constraint_set()
    rows = []
    for con in cs.get_all_constraints():
        rows.append(get_constraint_score_data(pose, con))
    return pd.DataFrame(rows)


def print_bad_constraint_scores(pose, cutoff=0.5):
    cs = pose.constraint_set()
    for con in cs.get_all_constraints():
        data = get_constraint_score_data(pose, con)
        if data['score'] > cutoff:
            print('{constraint_name}: {atom_A} – {atom_B} {function_name}={funpart} --> {score:.2}'.format(**data))


def get_AtomID(pose: pyrosetta.Pose, chain: str, resi: int, atomname: str) -> pyrosetta.AtomID:
    r = pose.pdb_info().pdb2pose(res=resi, chain=chain)
    assert r != 0, f'{resi}:{chain} is absent'
    residue = pose.residue(r)
    return pyrosetta.AtomID(atomno_in=residue.atom_index(atomname), rsd_in=r)


def get_AtomID_by_NGL_sele(pose: pyrosetta.Pose, selection: str) -> pyrosetta.AtomID:
    """
    23:A.CA
    """
    if ' ' in selection.strip():
        raise ValueError('single atom selection')
    # chain
    if ':' in selection:
        chain = re.match(r':(\w)', selection).group(1)
    else:
        chain = 'A'
    # atom name
    if ':' in selection:
        name = re.match(r'\.(\w+)', selection).group(1)
    else:
        name = ''
    # residue name
    if '[' in selection:
        resn = re.match(r'\[(\w+)]', selection).group(1)
    else:
        resn = ''
    # residue index
    if re.search(r'\d', selection):
        # sub pattern was incorrect, '' added, but I am not sure what it does.
        resi = int(re.match(r'(\d+)', re.sub(r'\[.*]', '', selection)).group(1))
    else:
        resi = float('nan')
    # assert
    if str(resi) == 'nan' or chain == '' or name == '':
        raise ValueError(f'selection {selection} is not like `23:A.CA`.')
    # return
    return get_AtomID(pose, chain=chain, resi=resi, atomname=name)


def get_AtomID_from_pymol_line(pose: pyrosetta.Pose, line: Optional[str] = None) -> pyrosetta.rosetta.core.id.AtomID:
    """
    Given a copypaste from the console in pymol following an atom selection in edit mode:
    (``You clicked /1amq/B/A/PMP`413/N1 -> (pk2)``) returns that atom in PyRosetta.

    If the line argument is blank the clipboard is read.

    :param pose:
    :param line:
    :return:
    """
    # You clicked /1amq/B/A/PMP`413/N1 -> (pk2)
    if line is None:
        import xerox
        line = xerox.paste()
    rex = re.search(r'\/\w+/\w?/(\w)/\w{3}`(\d+)/(\w+) ', line)
    chain, res, atomname = rex.groups()
    return get_AtomID(pose, chain, int(res), atomname)


def make_constraint_from_pymol_line(pose: pyrosetta.Pose,
                                    lines: str) \
        -> pyrosetta.rosetta.core.scoring.constraints.AtomPairConstraint:
    """
    two atoms clicked...
    You clicked /1amq/A/A/ASP`222/OD2 -> (pk1)
    You clicked /1amq/B/A/PMP`413/N1 -> (pk2)
    distance measured in PyRosetta hence the pose.

    :param lines:
    :param pose:
    :return:
    """
    fore, aft = lines.strip().split('\n')
    HarmonicFunc = pyrosetta.rosetta.core.scoring.func.HarmonicFunc
    AtomPairConstraint = pyrosetta.rosetta.core.scoring.constraints.AtomPairConstraint
    fore_atom = get_AtomID_from_pymol_line(pose, fore)
    aft_atom = get_AtomID_from_pymol_line(pose, aft)
    fore_xyz = pose.residue(fore_atom.rsd()).xyz(fore_atom.atomno())
    aft_xyz = pose.residue(aft_atom.rsd()).xyz(aft_atom.atomno())
    d = (fore_xyz - aft_xyz).norm()
    return AtomPairConstraint(fore_atom, aft_atom, HarmonicFunc(x0_in=d, sd_in=0.2))



"""
There is likely a mover that does this already and better...
"""


def make_ss(pose, begin: int = 1, end: int = -1, phi: float = 180, psi: float = 180):
    """
    Given a pose, likely one made via ``pyrosetta.pose_from_sequence``,
    make the fortran-indexed range [begin, end] (inclusive) into that SS.
    """
    if end == -1:
        end = pose.total_residue()
    for r in range(begin, end + 1):
        pose.set_phi(r, phi)
        pose.set_psi(r, psi)


def make_alpha_helical(pose, begin: int = 1, end: int = -1):
    make_ss(pose, begin, end, phi=-57.8, psi=-47.0)


def make_310_helical(pose, begin: int = 1, end: int = -1):
    make_ss(pose, begin, end, phi=-74.0, psi=-4.0)


def make_pi_helical(pose, begin: int = 1, end: int = -1):
    make_ss(pose, begin, end, phi=-57.1, psi=-69.7)


def make_sheet(pose, begin: int = 1, end: int = -1):
    make_ss(pose, begin, end, phi=-139, psi=+135)

# PyCharm is licking a mandrake again. Ignore the warning of bytes + str
make_alpha_helical.__doc__ = make_ss.__doc__ + '\nIn this case alpha-helix with phi=-57.8, psi=-47.0'
make_310_helical.__doc__ = make_ss.__doc__ + '\nIn this case 3.10-helix with phi=-74.0, psi=-4.0'
make_pi_helical.__doc__ = make_ss.__doc__ + '\nIn this case pi-helix with phi=-57.1, psi=-69.7'
make_sheet.__doc__ = make_ss.__doc__ + '\nIn this case sheet with phi=-139, psi=+135'



import json
from typing import (List)


class ChainOps:
    """
    Works on the list of dict (metadata.json) with among others keys

    * number (pose number)
    * chain (chain letter)
    * gene_name (gene name)

    """

    def __init__(self, chains: List[dict]):
        self.chains = chains

    def load(self, json_filename: str = 'metadata.json'):
        with open(json_filename, 'r') as fh:
            self.chains = json.load(fh)

    def dump(self, json_filename: str = 'metadata.json'):
        with open(json_filename, 'w') as fh:
            json.dump(self.chains, fh)

    def get_entry_of_key(self, value, key):
        return [chain for chain in self.chains if chain[key] == value][0]

    def get_entry(self, value):
        """
        guess key...

        :param value:
        :return:
        """
        if isinstance(value, int):
            return self.get_entry_of_key(value, 'number')
        elif isinstance(value, str) and len(value) == 1:
            return self.get_entry_of_key(value, 'chain')
        elif isinstance(value, str):
            return self.get_entry_of_key(value, 'gene_name')
        elif isinstance(value, dict) and 'gene_name' in value:  # an entry was passed
            return value

    def get_pose_of_chain(self, pose, value):  # pose is a pyrosetta.Pose
        return pose.split_by_chain()[self.get_entry(value)['number']]

    def __getitem__(self, value):
        return self.get_entry(value)

    def __iter__(self):
        return self.chains



__all__ = ['RingSelector',
           'AlteredSelector',
           'UnalteredSelector',
           'OrListSelector']

# will be fixed.
__all__.append('get_bfactor_vector')

import pyrosetta


class RingSelector:
    """
    Select all residues in the "ring"
    based upon 12A from origin.
    There is probably a saner way.

    NB> This is not actually a residue selector. The logical selectors will not accept it.
    """

    def __init__(self, radius=12):
        self.radius = radius

    def apply(self, pose: pyrosetta.Pose) -> pyrosetta.rosetta.utility.vector1_bool:
        sele = pyrosetta.rosetta.utility.vector1_bool(pose.total_residue())
        for r in range(1, pose.total_residue() + 1):
            if abs(pose.residue(r).xyz(1).x) < 12:
                sele[r] = 1
        return sele


class AlteredSelector:
    """
    Select residues that were altered in the threading.
    NB> This is not actually a residue selector. The logical selectors will not accept it.
    """

    def __init__(self, threader: pyrosetta.rosetta.protocols.comparative_modeling.ThreadingMover):
        self.threader = threader

    def apply(self, pose: pyrosetta.Pose) -> pyrosetta.rosetta.utility.vector1_bool:
        sele = pyrosetta.rosetta.utility.vector1_bool(pose.total_residue())
        mapping = self.threader.get_qt_mapping(pose).mapping()
        for r in range(1, pose.total_residue() + 1):
            if mapping[r] != 0:
                sele[r] = 1
        return sele


class UnalteredSelector:
    """
    Select residues that were unaltered in the threading.
    NB> This is not actually a residue selector. The logical selectors will not accept it.
    """

    def __init__(self, threader: pyrosetta.rosetta.protocols.comparative_modeling.ThreadingMover):
        self.threader = threader

    def apply(self, pose: pyrosetta.Pose) -> pyrosetta.rosetta.utility.vector1_bool:
        sele = pyrosetta.rosetta.utility.vector1_bool(pose.total_residue())
        mapping = self.threader.get_qt_mapping(pose).mapping()
        for r in range(1, pose.total_residue() + 1):
            if mapping[r] == 0:
                sele[r] = 1
        return sele


def OrListSelector(*selectors) -> pyrosetta.rosetta.core.select.residue_selector.OrResidueSelector:
    """
    OrResidueSelector but 2+
    (not a class, but returns a Or
    :param selectors:
    :return:
    """
    sele = pyrosetta.rosetta.core.select.residue_selector.FalseResidueSelector()
    for subsele in selectors:
        sele = pyrosetta.rosetta.core.select.residue_selector.OrResidueSelector(subsele, sele)
    return sele


# --------------------------------------------------------------------------------------------
# not a a faux selector

def get_bfactor_vector(pose: pyrosetta.Pose, cutoff: float, above=True) -> pyrosetta.rosetta.utility.vector1_bool:
    """
    Return a selection vector based on b-factors.
    above = get all above. So to select bad b-factors above is ``True``,
    but to select AF2 bad ones. above is ``False``
    """
    pdb_info = pose.pdb_info()
    vector = pyrosetta.rosetta.utility.vector1_bool(pose.total_residue())
    for r in range(1, pose.total_residue() + 1):
        try:
            atom_index = pose.residue(r).atom_index('CA')
        except AttributeError:
            atom_index = 1
        bfactor = pdb_info.bfactor(r, atom_index)
        if above and bfactor >= cutoff:
            vector[r] = True
        elif not above and bfactor <= cutoff:
            vector[r] = True
        else:
            pass  # vector[r] = False
    return vector



__all__ = ['add_pae_constraints',
           'add_interchain_pae_constraints',
           'add_stretch_constraint',
           'get_distance_matrix',
           'make_pae_constraint']

from typing import (Optional)

import numpy as np
import pyrosetta


def add_pae_constraints(pose: pyrosetta.Pose,
                        errors: np.ndarray,
                        cutoff: float = 12,
                        tolerance: Optional[float] = None,
                        weight: float = 1,
                        adjecency_threshold=5) -> None:
    """

    Add constrains to the pose based on the errors matrix.
    NB. this matrix is a reshaped version of what AF2 returns.

    A harmonic function is added to CA atoms that are in residues with the error under a specified cutoff.
    The mu is the current distance and the standard deviation of the harmonic is the error times ``weight``.

    To find out how many were added:

    >>> len(pose.constraint_set().get_all_constraints())

    :param pose:
    :param errors:
    :param cutoff:
    :param tolerance: if None Harmonic, if value, tollerance of FlatHarmonic
    :param weight: this is added to the SD part so squared inverse.
    :param adjecency_threshold: min residue separation of sequence neighbours
    :return:
    """
    for r1_idx, r2_idx in np.argwhere(errors < cutoff):
        if abs(r1_idx - r2_idx) < adjecency_threshold:
            continue # skip neighbours
        elif r1_idx <= r2_idx:
            continue # add once.
        d_error = errors[r1_idx, r2_idx]
        apc = make_pae_constraint(pose=pose,
                                  residue1_pose_idx=r1_idx + 1,
                                  residue2_pose_idx=r2_idx + 1,
                                  error=d_error,
                                  weight=weight,
                                  tolerance=tolerance)
        pose.add_constraint(apc)

def make_pae_constraint(pose,
                        residue1_pose_idx:int, # one indexed.
                        residue2_pose_idx:int, # one indexed.
                        error:float,
                        tolerance: Optional[float] = None,
                        weight:float=1):
    get_ca = lambda r, i: pyrosetta.AtomID(atomno_in=r.atom_index('CA'), rsd_in=i)
    FlatHarmonicFunc = pyrosetta.rosetta.core.scoring.func.FlatHarmonicFunc
    HarmonicFunc = pyrosetta.rosetta.core.scoring.func.HarmonicFunc

    AtomPairConstraint = pyrosetta.rosetta.core.scoring.constraints.AtomPairConstraint
    residue1 = pose.residue(residue1_pose_idx)
    ca1_atom = get_ca(residue1, residue1_pose_idx)
    residue2 = pose.residue(residue2_pose_idx)
    ca2_atom = get_ca(residue2, residue2_pose_idx)
    ca1_xyz = residue1.xyz(ca1_atom.atomno())
    ca2_xyz = residue2.xyz(ca2_atom.atomno())
    d = (ca1_xyz - ca2_xyz).norm()
    if not tolerance:
        fun = HarmonicFunc(x0_in=d, sd_in=error * weight)
    else:
        fun = FlatHarmonicFunc(x0_in=d, sd_in=error * weight, tol_in=tolerance)
    return AtomPairConstraint(ca1_atom, ca2_atom, fun)


def add_interchain_pae_constraints(pose, errors, cutoff=15):
    xdistances = get_distance_matrix(pose)
    for c in (1, 2):
        xdistances[pose.chain_begin(c) - 1: pose.chain_end(c),
        pose.chain_begin(c) - 1: pose.chain_end(c)] = np.nan
    with np.errstate(invalid='ignore'):
        mask = xdistances < cutoff
    for r1_idx, r2_idx in np.argwhere(mask):
        apc = make_pae_constraint(pose=pose,
                                  residue1_pose_idx=r1_idx + 1,
                                  residue2_pose_idx=r2_idx + 1,
                                  error=errors[r1_idx, r2_idx],
                                  tolerance=3)  # flatharmonic
        pose.add_constraint(apc)


def add_stretch_constraint(pose: pyrosetta.Pose,
                           weight: float = 5,
                           slope_in: float = -0.05,
                           residue_index_A: int = 1,
                           residue_index_B: int = -1,
                           distance: Optional[float] = None,
                           sigmoid: bool = True
                           ) -> pyrosetta.rosetta.core.scoring.constraints.AtomPairConstraint:
    """
    Add a constraint to "stretch out" the model, because ``slope_in`` is negative.

    :param pose: Pose to add constraint to
    :param weight: how strength of constraint (max of 0.5 for ``SigmoidFunc``)
    :param slope_in: negative number to stretch
    :param residue_index_A: first residue?
    :param residue_index_B: last residue is "-1"
    :param distance: if omitted, the midpoint of Sigmoid will be the current distance
    :param sigmoid: use sigmoid or identity/linear (bad idea)
    :return:
    """
    # get current length
    first_ca = pyrosetta.AtomID(atomno_in=pose.residue(residue_index_A).atom_index('CA'),
                                rsd_in=residue_index_A)
    if residue_index_B == -1:
        residue_index_B = pose.total_residue()
    last_ca = pyrosetta.AtomID(atomno_in=pose.residue(residue_index_B).atom_index('CA'),
                               rsd_in=residue_index_B)
    first_ca_xyz = pose.residue(1).xyz(first_ca.atomno())
    last_ca_xyz = pose.residue(pose.total_residue()).xyz(last_ca.atomno())
    if distance is None:
        distance = (first_ca_xyz - last_ca_xyz).norm()
    # make & add con
    sf = pyrosetta.rosetta.core.scoring.func
    AtomPairConstraint = pyrosetta.rosetta.core.scoring.constraints.AtomPairConstraint
    if sigmoid:
        fun = sf.ScalarWeightedFunc(weight, sf.SigmoidFunc(x0_in=distance, slope_in=slope_in))
    else:
        fun = sf.ScalarWeightedFunc(weight/distance, sf.IdentityFunc())
    con = AtomPairConstraint(first_ca, last_ca, fun)
    pose.add_constraint(con)
    return con

def get_distance_matrix(pose) -> np.ndarray:
    """
    Note the distance matrix is zero indexed as it would be confusing using numpy with one indexed data.

    :param pose:
    :return:
    """
    distances = np.zeros((pose.total_residue(), pose.total_residue()))
    ca_xyzs = [pose.residue(r).xyz('CA') for r in range(1, pose.total_residue() + 1)]
    for i in range(len(ca_xyzs)):
        for j in range(i):
            d = ca_xyzs[i].distance(ca_xyzs[j])
            distances[i, j] = d
            distances[j, i] = d
    return distances



class BlueprinterCommon:
    def pick_native(self, index):
        """
        NATAA and PIKAA native have different results... the latter is better.
        Does nothing if index does not exist.
        """
        # this is not a typo. get item returns the row, setter sets the value (index 3)
        try:
            self[index] = f'PIKAA {self[index][1]}'
        except ValueError:
            pass

    def del_span(self, start: int, stop: int):
        self.pick_native(start - 1)
        del self[start:stop]
        self.pick_native(start + 1)

    def wobble_span(self, start: int, stop: int):
        for resi in range(start,stop+1):
            self.pick_native(resi)

    def mutate(self, resi: int, to_resn: str):
        self.pick_native(resi - 1)
        self[resi] = f'PIKAA {to_resn}'
        self.pick_native(resi + 1)

    def expand_loop_wobble(self):
        raise NotImplementedError
        # identify loops
        loop_resis = [i+1 for i, s in enumerate(self.ss) if s in ('L', 'D')]
        # escape if pointless
        if len(loop_resis) == 0:
            return
        # loops to tuple ranges
        start = loop_resis[0]
        stop = loop_resis[0]
        loop_resi_ranges = []
        for resi in loop_resis:
            if resi != start + 1:
                loop_resi_ranges.append((start, stop))
                start = resi
            stop = resi
        else:
            loop_resi_ranges.append((start, stop))
        # filter ranges that contain mutations
        # [4, 'I', 'H', 'PIKAA I']
        expandenda = []
        for row in self.rows:
            # [4, 'I', 'H', 'PIKAA I']
            if row[2] in ('L', 'D'):
                for start, stop in loop_resi_ranges:
                    if start <= row[0] <= stop:
                        expandenda.append((start, stop))
        # expand ranges
        for start, stop in expandenda:
            for resi in range(start, stop+1):
                try:
                    if self[resi][2] == '.':
                        self.pick_native(resi)
                except:
                    pass




# mixin classes are private
from ._init import BlueprinterInit as _Init
from ._common import BlueprinterCommon as _Common
from ._subscripted import BlueprinterSubscripted as _Subscripted
from ._expected import BlueprinterExpected as _Expected
from ._remodel import Remodel as _Remodel
from ._pdb_info import BlueCopier as _Copier
# public
from ._pdb_info import ResInfo


class Blueprinter(_Init, _Subscripted, _Common, _Expected, _Remodel, _Copier):
    """
    Make a blueprint file for Rosetta Remodel and load it into the options (``.set(fn)``).
    The rows property is a list of lists, this is what the operations manipulate.
    Where each row is something like [20, 'G', '.']
    subscript assignment allows the 4th field to be altered. The SS is based on the pose.

    * ``.from_pose(pose)`` initialises it from pose, while the reg. init is from sequence and ss.

    Whereas

    * ``.del_span(10, 29)`` deletes a span
    * ``.wobble_span(100, 105)`` wobbles a span (NATAA)
    * ``.mutate(5, 'W')`` mutates resi i to aa
    * ``.insert(3, 'PIKAA W')`` or ``.insert(3, ['PIKAA W', 'PIKAA A'])`` inserts.

    a more flexible approach is using subscripts and slices:

    >>> blue = Blueprinter.from_pose(pose)
    >>> blue[10:14] = 'NATAA' # preceding loop
    >>> del blue[15:20]
    >>> blue[20:25] = 'NATAA' # following loop
    >>> blue[22] = 'PIKAA W'

    blue.set(bluprint_filename)

    remember that if remodelmover is already initialised to call ``rm.register_options()``

    About the subscript setter, there are three (weird) things to note:

    * A star will be convered into the native amino acid. `PIKAA F*CK` on a glutamine, will result in `PIKAA FECK`
    * The ranges are human style: deletion of resi 10-15 means that 6 residues are missing, including 15. Where normally in python 10:15 is 5, 15 is excluded.
    * The value is a string but there are no safegards or checks, even if there are a limited amount of possibilities and it would be really nice seeing if the non-canonical for `EMPTY NC XAA`
    """
    ResInfo = ResInfo

    # ========= init ===========================
    # from BlueprinterInit
    # adding init and classmethods

    # ========= subscripted ===========================
    # from BlueprinterSubscripted
    # adds subscription methods

    # === common operations =============================================
    # from BlueprinterCommon
    # adds some common tasks

    # === expected  =============================================
    # from BlueprinterExpected
    # adds expected_seq dynamic property




import os
from typing import (Optional)

import pyrosetta

__all__ = ['get_local_scorefxn',
           'prep_ED',
           'get_local_relax',
           'do_local_relax',
           'do_chainwise_relax']

def get_local_scorefxn() -> pyrosetta.ScoreFunction:
    ## local scorefxn w/ ED
    # ---------- Local weights ------------------
    # these are mostly the same except for the line:
    # <Set scale_sc_dens_byres="E:0.56,D:0.56,R:0.76,K:0.76,M:0.76,C:0.81,Q:0.81,H:0.81,N:0.81,T:0.81,S:0.81,
    # Y:0.88,W:0.88,A:0.88,F:0.88,P:0.88,I:0.88,L:0.88,V:0.88"/>
    # which is an utter mystery.
    weights = {"cart_bonded_length": 0.5,
               "cart_bonded_torsion": 0.5,
               "cart_bonded_angle": 1.0,
               "pro_close": 0.0,
               "fa_sol": 0.0,
               "elec_dens_fast": 30,  # <-- ED
               "rama": 0.0,
               "rama_prepro": 1.0}
    scorefxn_local = pyrosetta.create_score_function('ref2015_cart')
    stm = pyrosetta.rosetta.core.scoring.ScoreTypeManager()
    for name, value in weights.items():
        scorefxn_local.set_weight(stm.score_type_from_name(name), value)
    return scorefxn_local

def prep_ED(pose: pyrosetta.Pose, map_filename: str) -> pyrosetta.rosetta.core.scoring.electron_density.ElectronDensity:
    # rmsd & ED fit
    if os.path.splitext(map_filename) == '.gz':
        raise Exception('The file is gzipped')
    rmsd = pyrosetta.rosetta.core.simple_metrics.metrics.RMSDMetric(pose)
    ED = pyrosetta.rosetta.core.scoring.electron_density.getDensityMap(map_filename)
    initial_fit = ED.matchPose(pose)
    # This is redundant with ED.
    map_mover = pyrosetta.rosetta.protocols.cryst.LoadDensityMapMover(map_filename)
    map_mover.apply(pose)
    # This is redundant with map
    sdsm = pyrosetta.rosetta.protocols.electron_density.SetupForDensityScoringMover()
    sdsm.apply(pose)
    return ED


def get_local_relax(scorefxn: Optional[pyrosetta.ScoreFunction] = None,
                    ncyc: int = 3,
                    nexp: int = 2) -> pyrosetta.rosetta.protocols.relax.LocalRelax:
    if scorefxn is None:
        scorefxn = get_local_scorefxn()
    # <LocalRelax name="local_rlx" scorefxn="dens" max_iter="100" ncyc="1" ramp_cart="0" K="16" nexp="2"/>
    relax = pyrosetta.rosetta.protocols.relax.LocalRelax()
    relax.set_sfxn(scorefxn)
    relax.set_K(16)
    relax.set_max_iter(100)
    relax.set_ncyc(ncyc)
    relax.set_nexp(nexp)
    return relax

def do_local_relax(pose: pyrosetta.Pose, scorefxn: Optional[pyrosetta.ScoreFunction] = None) -> None:
    get_local_relax(scorefxn).apply(pose)


def do_chainwise_relax(pose: pyrosetta.Pose,
                       scorefxn: Optional[pyrosetta.ScoreFunction] = None,
                       cycles: int = 5) -> None:
    if scorefxn is None:
        scorefxn = pyrosetta.get_fa_scorefxn()
    for chain_i in range(1, pose.num_chains() + 1):
        chain_sele = pyrosetta.rosetta.core.select.residue_selector.ChainSelector(chain_i)
        chain_vector = chain_sele.apply(pose)
        movemap = pyrosetta.MoveMap()
        movemap.set_bb(allow_bb=chain_vector)
        movemap.set_chi(allow_chi=chain_vector)
        relax = pyrosetta.rosetta.protocols.relax.FastRelax(scorefxn, cycles)
        relax.set_movemap(movemap)
        relax.apply(pose)


from collections import Counter
from io import StringIO
from typing import (Tuple, Dict, List)

import pandas as pd
import requests
from Bio import SearchIO
from Bio.Blast.NCBIWWW import qblast


class LigandHunter:
    """
    Given a sequence find homologues (Blast) and find if they have ligands (PDBe API).
    The definition of what ligand is a cofactor comes from PDBe and does not count ions or
    triphospho-nucleotides as cofactors, but does count NADH adducts.

    * ``.data`` is a list of dictionaries.
    * ``.to_dataframe()`` converts into a pandas dataframe for further analysis.
    * ``.candidate_ligands`` list of ligand residue 3-letter codes
    """
    _cofactor_reference = {}

    @property
    def cofactor_reference(self) -> Dict[str, List[dict]]:
        if not self._cofactor_reference: #cached
            self._cofactor_reference = requests.get('https://www.ebi.ac.uk/pdbe/api/pdb/compound/cofactors').json()
        return self._cofactor_reference

    missing_codes = ['ATP', 'GTP', 'CA', 'MG', 'W']

    @property
    def cofactor_codes(self):
        _grouped_cofactor_codes = {name: value[0]['cofactors'] for name, value in self.cofactor_reference.items()}
        return [code for codes in _grouped_cofactor_codes.values() for code in codes] + self.missing_codes

    def __init__(self, sequence: str):
        """

        :param sequence: is assumed clean protein sequence
        """
        self.sequence = sequence
        self._blast_result = self._retrieve_homologues()  #:StringIO
        self.data = self._parse_blast_result()
        self._retrieve_ligands()  # inplace.
        self._ligand_data = None

    def to_dataframe(self):
        """
        Converts ``.data`` to a pandas dataframe
        """
        return pd.DataFrame(self.data).transpose()

    @property
    def candidate_ligands(self) -> List[str]:
        return list(set([code for datum in self.data.values() for code in datum['ligand_codes']]))

    @property  # cached the old way
    def ligand_data(self):
        """
        Data for the ligands.

        Note that LigandNicker has a get smiles method, that works like this, but is unrelated.
        """
        if self._ligand_data is None:
            self._ligand_data = requests.post('https://www.ebi.ac.uk/pdbe/api/pdb/compound/summary/',
                                              data=','.join(self.candidate_ligands)).json()
        return self._ligand_data

    def get_most_common_ligands(self) -> List[Tuple[str, int]]:
        # Uses collections.Counter not typing.Counter:
        c = Counter([code for datum in self.data.values() for code in datum['ligand_codes']])  # noqa
        return c.most_common()

    def get_pdb_entry_by_ligand(self, ligand_code: str) -> dict:
        """
        get pdb **entry** by ligand.
        Returns the first, which should be lowest e-value
        """
        for datum in self.data.values():
            if ligand_code in datum['ligand_codes']:
                return datum
        else:
            raise ValueError(f'{ligand_code} not found in any of the {len(self.data)} hits')

    # ------------ initialisation methods ------------------------

    def _retrieve_homologues(self) -> StringIO:
        return qblast('blastp', 'pdb', self.sequence)

    def _parse_blast_result(self) -> Dict[str, dict]:
        results = {}
        self._blast_result.seek(0)
        for query in SearchIO.parse(self._blast_result, 'blast-xml'):
            for hit in query:
                datum = dict(accession=hit.accession,
                             description=hit.description,
                             evalue=hit.hsps[0].evalue)
                datum['pdb_code'], datum['chain'] = hit.accession.split('_')
                results[datum['pdb_code'].upper()] = datum
        return results

    def _retrieve_ligands(self) -> None:
        query = ','.join(self.data.keys())
        reply = requests.post(url='https://www.ebi.ac.uk/pdbe/api/pdb/entry/ligand_monomers/',
                              data=query)
        for code, datum in reply.json().items():
            entry = self.data[code.upper()]
            entry['ligands'] = datum
            entry['ligand_codes'] = [inner['chem_comp_id'] for inner in datum]
            entry['cofactor_codes'] = [code for code in entry['ligand_codes'] if code in self.cofactor_codes]
            entry['has_cofactor'] = bool(entry['cofactor_codes'])



import pyrosetta

class BlueprinterInit:
    def __init__(self, seq: str, ss: str = None):
        self.seq = seq
        self.ss = ss
        self.rows = self._seq2rows(seq)

    def _seq2rows(self, seq):
        rows = []
        for i, aa in enumerate(seq):
            rows.append([i + 1, aa, '.'])
        return rows

    @classmethod
    def from_pose(cls, pose: pyrosetta.Pose):
        seq = pose.chain_sequence(1)
        DSSP = pyrosetta.rosetta.protocols.moves.DsspMover()
        DSSP.apply(pose)
        ss = pose.secstruct()  # for this case, no slicing.
        return cls(seq, ss)

    def get_ss(self, idx: int):
        if self.ss is not None and len(self.ss) > idx:
            return self.ss[idx - 1]
        else:
            return 'D'


from .weights import WeightWatcher
from .blueprint_maker import Blueprinter
from .chain_ops import ChainOps, Transmogrifier, Murinizer
from .score_mutants import Mutation, MutantScorer, extend_scores
from .common_ops import *  # common_ops  has __all__ in some files
from .init_ops import *  # init_ops.__init__ imports only needed functions
from .threading import *  # threading has __all__ in __init__
from .alphafold import *  # alphafold has __all__ in some files
from .ligands import *
from .per_atom import AtomicInteractions
from .residue_decription import *  # __init__ imports only needed functions








import pyrosetta

def get_ss(pose: pyrosetta.Pose) -> str:
    """
    returns a string of the SS of types H, S, L
    This is not the ALBEGO classifier
    """
    DSSP = pyrosetta.rosetta.protocols.moves.DsspMover()  # noqa
    DSSP.apply(pose)
    return pose.secstruct()


__all__ = ['download_map',
           'download_cif',
           'download_pdb',
           'download_opm']

import requests
import shutil
import urllib.request as request
from contextlib import closing

def download_map(code: str):
    """
    This is two functions. EMD database for cryoEM files

    :param code:
    :return:
    """
    if isinstance(code, int) or "EMD-" in code or code.isdigit():
        return _download_mrc_map(str(code))
    else:
        return _download_ccp4_map(code)

def _download_mrc_map(code:str):
    # alt:
    # https://ftp.wwpdb.org/pub/emdb/structures/EMD-20808/map/emd_20808.map.gz
    code = code.replace("EMD-", "")
    ftp_path = f'ftp://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-{code}/map/emd_{code}.map.gz'
    file_path = f'EMD-{code}.map.gz'
    with closing(request.urlopen(ftp_path)) as r, open(file_path, 'wb') as f:
        shutil.copyfileobj(r, f)
    return file_path

def _download_ccp4_map(code:str):
    code = code.replace("pdb:", "")
    assert len(code) == 4, f'PDB code {code} is not 4 alphanumericals long'
    url = f'https://www.ebi.ac.uk/pdbe/coordinates/files/{code.lower()}.ccp4'
    file_path = f'{code}.ccp4'
    reply = requests.get(url, stream=True)
    assert reply.status_code == 200
    with open(file_path, 'wb') as f:
        reply.raw.decode_content = True
        shutil.copyfileobj(reply.raw, f)
    return file_path

def download_cif(code: str):
    """
    Download CIF. Pyrosetta has issues importing Cifs (hetatms).
    This is just a note to self as PyMOL fetch can do it.
    """
    http_path = f'http://www.ebi.ac.uk/pdbe/static/entry/download/{code.lower()}-assembly-1.cif.gz'
    file_path = f'{code}.cif.gz'
    r = requests.get(http_path, stream=True)
    if r.status_code == 200:
        with open(file_path, 'wb') as f:
            r.raw.decode_content = True
            shutil.copyfileobj(r.raw, f)
        return True
    else:
        return False

def download_pdb(pdbcode: str) -> str:
    filename = f'{pdbcode.lower()}.pdb'

    # r = requests.get(f'https://www.ebi.ac.uk/pdbe/coordinates/files/{pdbcode.lower()}.ccp4', stream=True)
    r = requests.get(f'https://www.ebi.ac.uk/pdbe/entry-files/download/pdb{pdbcode.lower()}.ent', stream=True)
    if r.status_code == 200:
        with open(filename, 'wb') as f:
            r.raw.decode_content = True
            shutil.copyfileobj(r.raw, f)
    return filename

def download_opm(code: str):
    """
    Download OPM PDB
    """
    import shutil, requests
    http_path = f'https://opm-assets.storage.googleapis.com/pdb/{code.lower()}.pdb'
    file_path = f'{code}_OMP.pdb'
    r = requests.get(http_path, stream=True)
    if r.status_code == 200:
        with open(file_path, 'wb') as f:
            r.raw.decode_content = True
            shutil.copyfileobj(r.raw, f)
        return True
    else:
        return False


import pyrosetta
import sys

if sys.version_info < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict
from typing import Dict, List
from ..common_ops import pose_range

BondDataType = TypedDict('BondDataType',
                         {'distance':     float,
                          'energy':       float,
                          'acc_resi':     int,
                          'acc_resn':     str,
                          'acc_atm_name': str,
                          'don_resi':     int,
                          'don_resn':     str,
                          'don_atm_name': str,
                          })

def get_hbond_dicts(pose: pyrosetta.Pose) -> Dict[int, List[BondDataType]]:
    hbondset = pose.get_hbonds()
    return {i: [hbond2dict(pose, bond) for bond in hbondset.residue_hbonds(i)] for i in pose_range(pose)}


def hbond2dict(pose: pyrosetta.Pose, bond: pyrosetta.rosetta.core.scoring.hbonds.HBond) -> BondDataType:
    return BondDataType(distance=bond.get_HAdist(pose),
                        energy=bond.energy(),
                        **_get_acceptor(pose, bond),
                        **_get_donor(pose, bond))


def _parse_part(pose: pyrosetta.Pose, resi: int, atom: int, prefix: str = '') -> dict:
    residue: pyrosetta.rosetta.core.conformation.Residue = pose.residue(resi)
    resn: str = residue.name3()
    atm_name: str = residue.atom_name(atom).strip()
    data = dict(resi=resi, resn=resn, atm_name=atm_name)
    return {f'{prefix}_{k}': v for k, v in data.items()} if prefix else data


def _get_acceptor(pose: pyrosetta.Pose, bond: pyrosetta.rosetta.core.scoring.hbonds.HBond) -> dict:
    return _parse_part(pose, resi=bond.acc_res(), atom=bond.acc_atm(), prefix='acc')


def _get_donor(pose: pyrosetta.Pose, bond: pyrosetta.rosetta.core.scoring.hbonds.HBond) -> dict:
    return _parse_part(pose, resi=bond.don_res(), atom=bond.don_hatm(), prefix='don')




# https://www.rosettacommons.org/docs/latest/rosetta_basics/scoring/score-types
term_meanings = {
    "fa_atr": "Lennard-Jones attractive between atoms in different residues (r^6 term, London dispersion forces).",
    "fa_rep": "Lennard-Jones repulsive between atoms in different residues (r^12 term, Pauli repulsion forces).",
    "fa_sol": "Lazaridis-Karplus solvation energy.",
    "fa_intra_rep": "Lennard-Jones repulsive between atoms in the same residue.",
    "fa_elec": "Coulombic electrostatic potential with a distance-dependent dielectric.",
    "pro_close": "Proline ring closure energy and energy of psi angle of preceding residue.",
    "hbond_sr_bb": "Backbone-backbone hbonds close in primary sequence.",
    "hbond_lr_bb": "Backbone-backbone hbonds distant in primary sequence.",
    "hbond_bb_sc": "Sidechain-backbone hydrogen bond energy.",
    "hbond_sc": "Sidechain-sidechain hydrogen bond energy.",
    "dslf_fa13": "Disulfide geometry potential.",
    "rama": "Ramachandran preferences.",
    "omega": "Omega dihedral in the backbone. A Harmonic constraint on planarity with standard deviation of ~6 deg.",
    "fa_dun": "Internal energy of sidechain rotamers as derived from Dunbrack's statistics (2010 Rotamer Library used in Talaris2013).",
    "fa_dun_semi": "Internal energy of sidechain semi-rotamers as derived from Dunbrack's statistics (2010 Rotamer Library used in Talaris2013).",
    "p_aa_pp": "Probability of amino acid at Φ/Ψ.",
    "ref": "Reference energy for each amino acid. Balances internal energy of amino acid terms.  Plays role in design.",
    "METHOD_WEIGHTS": "Not an energy term itself, but the parameters for each amino acid used by the ref energy term.",
    "lk_ball": "Anisotropic contribution to the solvation.",
    "lk_ball_iso": "Same as fa_sol; see below.",
    "lk_ball_wtd": "weighted sum of lk_ball & lk_ball_iso (w1*lk_ball + w2*lk_ball_iso); w2 is negative so that anisotropic contribution(lk_ball) replaces some portion of isotropic contribution (fa_sol=lk_ball_iso).",
    "lk_ball_bridge": "Bonus to solvation coming from bridging waters, measured by overlap of the 'balls' from two interacting polar atoms.",
    "lk_ball_bridge_uncpl": "Same as lk_ball_bridge, but the value is uncoupled with dGfree (i.e. constant bonus, whereas lk_ball_bridge is proportional to dGfree values).",
    "fa_intra_atr_xover4": "Intra-residue LJ attraction, counted for the atom-pairs beyond torsion-relationship.",
    "fa_intra_rep_xover4": "Intra-residue LJ repulsion, counted for the atom-pairs beyond torsion-relationship.",
    "fa_intra_sol_xover4": "Intra-residue LK solvation, counted for the atom-pairs beyond torsion-relationship.",
    "fa_intra_elec": "Intra-residue Coulombic interaction, counted for the atom-pairs beyond torsion-relationship.",
    "rama_prepro": "Backbone torsion preference term that takes into account of whether preceding amono acid is Proline or not.",
    "hxl_tors": "Sidechain hydroxyl group torsion preference for Ser/Thr/Tyr, supersedes yhh_planarity (that covers L- and D-Tyr only).",
    "yhh_planarity": "Sidechain hydroxyl group torsion preference for Tyr, superseded by hxl_tors"
}